/* Web enabled cheese heat controller for Snow Farm Creamery
 * William F. Dudley Jr.
 * 2017 09 12
 * revised 2018 11 04

    todo:

    nil

    pin usage:

    Analog  0
    Analog  1
    Analog  . . .
    Analog  7
    Analog  8
    Analog 10
    Analog 11
    Analog 12
    Analog 13
    Analog 14
    Analog 15

    Digital 0 Serial input from USB interface
    Digital 1 Serial output to USB interface
    Digital 4 SD card select on ethernet board
PINOUT
     2	Relay for heat control
    16
    12
    18,19 TX1,RX1
    20,21 SDA,SCL	I2C to LCDisplay
    24,25 SDA,SCL
    40,41,42,43		Four push buttons
 */
#include <Arduino.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/io.h>
#include <stdlib.h>
#include <EEPROM.h>
#include "EEPROM_rw_anything.h"
#include "dbg.h"
// #include <Time.h>
#include <SPI.h>
#define PROFILING 0
#define PROFILING_MAIN 1
#include "profiling.h"
#include "pitches.h"

#include "Adafruit_MAX31855.h"
// Example creating a thermocouple instance with software SPI on any three
// digital IO pins.
#define MAXDO   5
#define MAXCS   6
#define MAXCLK  7
// initialize the Thermocouple
Adafruit_MAX31855 thermocouple(MAXCLK, MAXCS, MAXDO);

// Example creating a thermocouple instance with hardware SPI
// on a given CS pin.
//#define MAXCS   10
//Adafruit_MAX31855 thermocouple(MAXCS);


#define IS_MAIN 1
#if IS_MAIN
    #define EXTERN
    #define INIT(x) =x
    #define INIT3(x,y,z) =x,y,z
#else
    #define EXTERN extern
    #define INIT(x)
    #define INIT3(x,y,z)
#endif

#define HEAT_RELAY 2
#define HEAT_RELAY_POLARITY 1

#define DISPROW 4
#define DISPCOL 20

// I2C LCD screen PCF8574T I2C LCD Backpack 
#include <Wire.h>
#include <LCD.h>
#include <LiquidCrystal_I2C.h>
#define BACKLIGHT_PIN  3
LiquidCrystal_I2C  lcd(0x27,2,1,0,4,5,6,7,BACKLIGHT_PIN,POSITIVE);
// some LCDs are at 0x3f, some are at 0x27

#define SEEED_ETHERNET 1
#if SEEED_ETHERNET
#include <EthernetV2_0.h>
#include <EthernetUdpV2_0.h>
#include <DnsV2_0.h>
#else
// #include <Ethernet.h>
// #include <EthernetUdp.h>
// #include <Dns.h>
#endif
#define PREFIX ""
#define W5200_CS  10
#define SDCARD_CS 4

#include "WebServer.h"
// Initialize the Ethernet client library
// with the IP address and port of the server
// that you want to connect to (port 80 is default for HTTP):

EthernetServer Control(80);
EthernetClient client;
void defaultCmd(WebServer &, WebServer::ConnectionType, char *, bool);
void prog_Cmd(WebServer &, WebServer::ConnectionType, char *, bool);
void netw_Cmd(WebServer &, WebServer::ConnectionType, char *, bool);
// prog_Cmd -> output_prog() -> show_prog_page() -> foreach program, show_template(server, type, program, programtmpl)
// netw_Cmd -> output_netw()
// sett_Cmd -> output_sett()
void output_graph (WebServer &, WebServer::ConnectionType, bool);
void setUIstate (uint8_t);
void save_log_to_eeprom (void);
void restore_log_from_eeprom (void);
void populate_log (void);

// no-cost stream operator as described at
// http://sundial.org/arduino/?page_id=119
template <class T> inline Print &operator <<(Print &obj, T arg) {
    obj.print(arg);
    return obj;
}

int Fvalue;		// temperature, filtered
float Ffvalue;
uint8_t trcount;	// count of temp readings since boot-up
uint8_t outlier_count;
uint8_t ramp_rate;
volatile int set_temp;	// temperature set point
volatile int set_temp_i;	// intermediate temperature set point
uint8_t repeat_key;	// not implemented, do I need this?
static char cbuf[80];
uint8_t ci;
uint8_t finished_run;

// PID is appropriate, as long as you assume the control value is
// "amount of time to turn the heater on"; treat control as really low
// frequency PWM

float integral;		// for PID
float last_error;	// for PID
uint8_t temp_reached;	// for BangBang
uint8_t heat_on;	// for BangBang
uint8_t ondiff_i, offdiff_i;

#define BUTTON_0 40
#define BUTTON_1 41
#define BUTTON_2 42
#define BUTTON_3 43

// UI states:
#include "ui_state.h"

long bufferthing;
uint8_t read_temp_now;
uint8_t current_program;
uint8_t final_program;
uint8_t prev_current_program;
volatile uint8_t phase, phase_ended, ramp_ended;
uint8_t phase_time;

#define MAXMSTIMERS 2
#define MAXSTIMERS 4
#define MAXMTIMERS 1

// min_Timers values:

// sec_Timers values:
#define LCD_UPDATE_TIMER 0
#define LCD_UPDATE_PER 5	// seconds
#define HEAT_TIMER 1
#define HEAT_RAMP_TIMER 2	// for N degrees in 5 minutes timing
#define KEY_REPEAT_TIMER 3	// not implemented, do I need this?

// msec_Timers values:
#define READ_TEMP 0
#define BLINK_LCD 1

#define TEMP_MSEC 250
#define BLINK_MSEC 433

uint8_t blink_state;

EXTERN unsigned long msec_Timers[MAXMSTIMERS];
EXTERN unsigned long sec_Timers[MAXSTIMERS];
EXTERN unsigned long min_Timers[MAXMTIMERS];

#define CONFIG_REVISION_LEVEL 12347L

#define MAX_NAME 16
#define MAX_PROGRAMS 10
#define MAX_PHASES 10

typedef struct temptime
{
    uint8_t temp;
    uint8_t ramp_rate;	// in degrees F per 5 minute interval
    uint8_t minutes;
} TEMPTIME;

typedef struct cprogram
{
    char name[MAX_NAME];
    TEMPTIME temp_min[MAX_PHASES];
} CPROGRAM;

typedef struct config_mem
{
    unsigned long magic_number;
    uint8_t last_program_used;
    uint8_t config_revision;
    uint8_t ip[4];
    uint8_t dns_server[4];
    uint8_t gateway_ip[4];
    uint8_t subnet_mask[4];
    uint8_t mac[6];
    CPROGRAM programs[MAX_PROGRAMS];
    int8_t ondiff[3], offdiff[3], use_pid;
    float kp, ki, kd;
    uint16_t pwm_period_sec;
    uint8_t min_pwm_sec;
    uint8_t heater_start_delay;
    uint8_t log_index, log_start;	// log_start not used.  Use 0
    float tempfudge;
} CONFIG_DATA;

CPROGRAM eprogram;
CONFIG_DATA config_copy;

#define KP config_copy.kp
#define KI config_copy.ki
#define KD config_copy.kd
#define PWM_PER_SEC (float)config_copy.pwm_period_sec
#define MIN_PWM_SEC (float)config_copy.min_pwm_sec
#define HEATER_START_DELAY (float)config_copy.heater_start_delay

char Dbuf2[250];
char Dbuf[250];
uint8_t dns_working;
volatile uint8_t lcd_needs_update;
#define InterruptOff  TIMSK2 &= ~(1<<TOIE2)
#define InterruptOn  TIMSK2 |= (1<<TOIE2)
volatile unsigned int int_counter;
volatile unsigned char seconds;
volatile unsigned char new_minute;
volatile unsigned long uptime_min;
unsigned int tcnt2; // used to store timer value
void printlnIP (byte *);

/* log file output format:
Run: acquiring temp : sp = 92 : spi = 92 : tr = 0 : state = 1 : temp = 85.12
 * info: set_temp, set_temp_i, temp_reached, UIstate, Fvalue
 */
typedef struct logentry
{
    uint8_t set_temp;
    uint8_t set_temp_i;
    uint8_t temp_reached;
    uint8_t UIstate;
    uint8_t tempF;
    uint16_t sec_Timer;
} LOGENTRY;
#define MAXLOGENTRY 256		// must be <= 256 if log_index is uint8_t !!!

LOGENTRY run_log[MAXLOGENTRY];

void store_log(uint8_t tempF) {
    if(config_copy.log_index == (MAXLOGENTRY - 1)) { return; } // full !
    run_log[config_copy.log_index].set_temp = set_temp;
    run_log[config_copy.log_index].set_temp_i = set_temp_i;
    run_log[config_copy.log_index].temp_reached = temp_reached;
    run_log[config_copy.log_index].UIstate = UIstate;
    run_log[config_copy.log_index].sec_Timer = (uint16_t)sec_Timers[HEAT_TIMER];
    run_log[config_copy.log_index++].tempF = tempF;
    // eeprom_save_state();	// save log_index
}

void printlnIP (byte * ipa) {
    for(byte i = 0 ; i < 4 ; i++) {
	Serial.print(ipa[i]);
	Serial.print((i==3) ? '\n' : '.');
    }
}

// Arduino runs at 16 Mhz, so we have 1000 Overflows per second...
// this ISR will get hit once a millisecond
ISR(TIMER2_OVF_vect) {

    int_counter++;
    if (int_counter == 655) {	// what the hell?  should be 1000
	seconds++;
	int_counter = 0;
	for(uint8_t ic = 0 ; ic < MAXSTIMERS ; ic++) {
	    if(sec_Timers[ic]) {
		if(!(ic == HEAT_TIMER && UIstate == PAUSE)) {
		    sec_Timers[ic]--;
		}
		if(!sec_Timers[ic] && ic == LCD_UPDATE_TIMER) {
		    lcd_needs_update++;
		    sec_Timers[LCD_UPDATE_TIMER] = LCD_UPDATE_PER;
		}
		if(!sec_Timers[ic] && ic == HEAT_TIMER) {
		    phase_ended = 1;
		}
		if(!sec_Timers[ic] && ic == HEAT_RAMP_TIMER) {
		    ramp_ended = 1;	// causes do PID to start HEAT_RAMP_TIMER
		}
#if 0
		if(!sec_Timers[ic] && ic == KEY_REPEAT_TIMER) {
		    repeat_key = 1;	// not implemented, do I need this?
		}
#endif
	    }
	}
	if(seconds == 60) {
	    new_minute = 1;
	    seconds = 0;
	    uptime_min++;
	}
    }
    for(byte i = 0 ; i < MAXMSTIMERS ; i++) {
	if(msec_Timers[i]) {
	    msec_Timers[i]--;
	    if(!msec_Timers[i] && i == READ_TEMP) {
		read_temp_now = 1;
		msec_Timers[READ_TEMP] = TEMP_MSEC;
	    }
	    if(!msec_Timers[i] && i == BLINK_LCD) {
		blink_state ^= 1;
		msec_Timers[BLINK_LCD] = BLINK_MSEC;
		if(UIstate == EDIT
		|| UIstate == BANG
		|| UIstate == SET1
		|| UIstate == PID
		|| UIstate == NEWP) {
		    lcd_needs_update = 1;
		}
	    }
	}
    }
#if PROFILING
    prof_array[prof_line]++;
#endif
    TCNT2 = tcnt2;  // reset the timer for next time
}

// Timer setup code borrowed from Sebastian Wallin
// http://popdevelop.com/2010/04/mastering-timer-interrupts-on-the-arduino/
// This was the best explanation I found on the web.

void setupTimer() {
  //Timer2 Settings:  Timer Prescaler /1024
  // First disable the timer overflow interrupt while we're configuring
  TIMSK2 &= ~(1<<TOIE2);
  // Configure timer2 in normal mode (pure counting, no PWM etc.)
  TCCR2A &= ~((1<<WGM21) | (1<<WGM20));
  // Select clock source: internal I/O clock
  ASSR &= ~(1<<AS2);
  // Disable Compare Match A interrupt enable (only want overflow)
  TIMSK2 &= ~(1<<OCIE2A);

  // Now configure the prescaler to CPU clock divided by 128
  TCCR2B |= (1<<CS22)  | (1<<CS20); // Set bits
  TCCR2B &= ~(1<<CS21);		// Clear bit

  /* We need to calculate a proper value to load the timer counter.
   * The following loads the value 131 into the Timer 2 counter register
   * The math behind this is:
   * (CPU frequency) / (prescaler value) = 125000 Hz = 8us.
   * (desired period) / 8us = 125.
   * MAX(uint8) - 125 = 131;
   */
  /* Save value globally for later reload in ISR */
  tcnt2 = 131;

  /* Finally load end enable the timer */
  TCNT2 = tcnt2;
  TIMSK2 |= (1<<TOIE2);
  sei();
}

void eeprom_restore_state(void) {
    EEPROM_readAnything(0, config_copy);
}

void eeprom_save_state(void) {
CONFIG_DATA temp_config_copy;
    EEPROM_readAnything(0, temp_config_copy);
    if(memcmp((const void *)&temp_config_copy, (const void *)&config_copy, sizeof(CONFIG_DATA)/sizeof(uint8_t))) {
	EEPROM_writeAnything(0, config_copy);
    }
}

void LCDprintIP (byte *ipa) {
    for(byte i = 0 ; i < 4 ; i++) {
	lcd.print((int)ipa[i]);
	if(i < 3) lcd.print('.');
    }
}

// line is 0 .. 3 for four line LCD
// assumes 20 character width
void LCDclearLine (byte *line) {
    lcd.setCursor(0, line);
    lcd.print(F("                    "));
    lcd.setCursor(0, line);
}

void show_constants (void) {
    if(config_copy.use_pid) {
	DVF("KP =", config_copy.kp);
	DVF(": KI =", config_copy.ki);
	DVF(": KD =", config_copy.kd);
	DVF(": delay =", HEATER_START_DELAY);
    }
    else {
	DVF("ondiff[0] =", config_copy.ondiff[0]);
	DVF(": offdiff[0] =", config_copy.offdiff[0]);
	DVF(": ondiff[1] =", config_copy.ondiff[1]);
	DVF(": offdiff[1] =", config_copy.offdiff[1]);
	DVF(": ondiff[2] =", config_copy.ondiff[2]);
	DVF(": offdiff[2] =", config_copy.offdiff[2]);
    }
    DVFn(": use pid =", config_copy.use_pid ? "YES" : "NO");
}

void setup (void) {
uint16_t i;
uint8_t network_failed;
    network_failed = 0;
    finished_run = 0;
#if MAXLOGENTRY < 256
    if(config_copy.log_index >= MAXLOGENTRY) {
	config_copy.log_index = MAXLOGENTRY - 1;
    }
#endif
    heat_on = 0;
    ci = 0;
    memset(cbuf, '\0', sizeof(cbuf)/sizeof(char));
    Serial.begin(115200);
    Serial.println(F("System Initialization"));
#if PROFILING
    PF(0);
    prof_has_dumped = 0;
    clear_profiling_data();
#endif
    repeat_key = 0;	// not implemented, do I need this?
    pinMode(SDCARD_CS,OUTPUT);
    digitalWrite(SDCARD_CS,HIGH);//Deselect the SD card
    pinMode(HEAT_RELAY, OUTPUT);	// Heat Relay
    digitalWrite(HEAT_RELAY, 1 ^ HEAT_RELAY_POLARITY);	// off
    i = EEPROM_readAnything(0, config_copy);
    DVFn("bytes read from EE =", i);
    DVFn("test_revision =", config_copy.magic_number);
    DVFn("CONFIG_REVISION_LEVEL =", CONFIG_REVISION_LEVEL);
    // get saved settings
    if(config_copy.magic_number != CONFIG_REVISION_LEVEL) {
	memset((void *)&config_copy.magic_number, '\0', sizeof(CONFIG_DATA));
	config_copy.magic_number = CONFIG_REVISION_LEVEL;
	config_copy.ip[0] = 192;	// 192.168.1.243
	config_copy.ip[1] = 168;
	config_copy.ip[2] = 254;
	config_copy.ip[3] =  28;
	config_copy.gateway_ip[0] = config_copy.ip[0];
	config_copy.gateway_ip[1] = config_copy.ip[1];
	config_copy.gateway_ip[2] = config_copy.ip[2];
	config_copy.gateway_ip[3] = 254;
	config_copy.dns_server[0] = 8;
	config_copy.dns_server[1] = 8;
	config_copy.dns_server[2] = 4;
	config_copy.dns_server[3] = 4;
	config_copy.subnet_mask[0] = 255;
	config_copy.subnet_mask[1] = 255;
	config_copy.subnet_mask[2] = 255;
	config_copy.subnet_mask[3] =   0;
	config_copy.mac[0] = 0xde;
	config_copy.mac[1] = 0xad;
	config_copy.mac[2] = 0xbe;
	config_copy.mac[3] = 0xef;
	config_copy.mac[4] = 0xfe;
	config_copy.mac[5] = 0xc3;
	// put standard cheese programs in config_copy

	strcpy(config_copy.programs[0].name, "Baby Swiss");
	config_copy.programs[0].temp_min[0].temp = 00;
	config_copy.programs[0].temp_min[0].minutes = 000;
	config_copy.programs[0].temp_min[1].temp = 00;
	config_copy.programs[0].temp_min[1].minutes = 00;
	config_copy.programs[0].temp_min[2].temp = 00;
	config_copy.programs[0].temp_min[2].minutes = 00;
	config_copy.programs[0].temp_min[3].temp = 0;

	strcpy(config_copy.programs[1].name, "Fetish");
	config_copy.programs[1].temp_min[0].temp = 00;
	config_copy.programs[1].temp_min[0].minutes = 00;
	config_copy.programs[1].temp_min[1].temp = 0;

	strcpy(config_copy.programs[2].name, "Dream Street");
	config_copy.programs[2].temp_min[0].temp = 00;
	config_copy.programs[2].temp_min[0].minutes = 00;
	config_copy.programs[2].temp_min[1].temp = 00;
	config_copy.programs[2].temp_min[1].minutes = 5;
	config_copy.programs[2].temp_min[2].temp = 00;
	config_copy.programs[2].temp_min[2].minutes = 00;
	config_copy.programs[2].temp_min[3].temp = 0;

	strcpy(config_copy.programs[3].name, "Tiara/Asiago");
	config_copy.programs[3].temp_min[0].temp = 00;
	config_copy.programs[3].temp_min[0].minutes = 00;
	config_copy.programs[3].temp_min[1].temp = 00;
	config_copy.programs[3].temp_min[1].ramp_rate = 2;
	config_copy.programs[3].temp_min[1].minutes = 00;
	config_copy.programs[3].temp_min[2].temp = 00;
	config_copy.programs[3].temp_min[2].ramp_rate = 4;
	config_copy.programs[3].temp_min[2].minutes = 1;
	config_copy.programs[2].temp_min[3].temp = 0;

	strcpy(config_copy.programs[4].name, "Test (4)");
	config_copy.programs[4].temp_min[0].temp = 92;
	config_copy.programs[4].temp_min[0].minutes = 30;

	config_copy.programs[4].temp_min[1].temp = 104;
	config_copy.programs[4].temp_min[1].ramp_rate = 2;
	config_copy.programs[4].temp_min[1].minutes = 30;

	config_copy.programs[4].temp_min[2].temp = 104;
	config_copy.programs[4].temp_min[2].ramp_rate = 0;
	config_copy.programs[4].temp_min[2].minutes = 10;
	
	config_copy.programs[4].temp_min[3].temp = 120;
	config_copy.programs[4].temp_min[3].ramp_rate = 4;
	config_copy.programs[4].temp_min[3].minutes = 20;

	config_copy.programs[4].temp_min[4].temp = 0;

	config_copy.offdiff[0] = 1;
	config_copy.ondiff[0]  = 2;
	config_copy.offdiff[1] = 0;
	config_copy.ondiff[1]  = 1;
	config_copy.offdiff[2] = -1;
	config_copy.ondiff[2]  = 0;
	config_copy.use_pid  = 0;
	config_copy.kp       = 45.0;	// 4.45 for my test rig
	config_copy.ki       = 0.0000;	// 0.0005 for my test rig
	config_copy.kd       = 0.0;	// not used as yet
	config_copy.pwm_period_sec = 300;	// five minutes
	config_copy.min_pwm_sec = 1;
	config_copy.heater_start_delay = 4;
	config_copy.tempfudge = 0.0;
	Serial.println(F("initialized CONFIG_DATA"));
	EEPROM_writeAnything(0, config_copy);
	if(CONFIG_REVISION_LEVEL == 12347L) {
	    populate_log();	// just for testing log feature
	}
    }
    else {
	restore_log_from_eeprom();
    }
    DVFn("before memset eprogram, size =", sizeof(CPROGRAM));
    memset((void *)&eprogram.name[0], '\0', sizeof(CPROGRAM));
    if(config_copy.programs[6].name[0]) {
	DVFn("before memcpy program 6, size =", sizeof(CPROGRAM));
	memcpy((void *)&config_copy.programs[6].name[0], (void *)&eprogram.name[0], sizeof(CPROGRAM));
    }
    if(config_copy.programs[7].name[0]) {
	DLn("before memcpy program 7");
	memcpy((void *)&config_copy.programs[7].name[0], (void *)&eprogram.name[0], sizeof(CPROGRAM));
    }
    if(config_copy.programs[8].name[0]) {
	DLn("before memcpy program 8");
	memcpy((void *)&config_copy.programs[8].name[0], (void *)&eprogram.name[0], sizeof(CPROGRAM));
    }
    if(config_copy.programs[9].name[0]) {
	DLn("before memcpy program 9");
	memcpy((void *)&config_copy.programs[9].name[0], (void *)&eprogram.name[0], sizeof(CPROGRAM));
    }
    DLn("after memcpy program 9");
    config_copy.use_pid = config_copy.use_pid ? 1 : 0;
    if(0) {
	config_copy.kp      = 45.0;	// 4.45 for my test rig
	config_copy.ki      = 0.0000;	// 0.0005 for my test rig
	config_copy.kd      = 0.0;	// not used as yet
    }
    if(config_copy.pwm_period_sec != 300) {
	config_copy.pwm_period_sec = 300;
    }
    if(config_copy.min_pwm_sec > 9) {
	config_copy.min_pwm_sec = 1;
    }
    if(config_copy.heater_start_delay > 10) {
	config_copy.heater_start_delay = 0;
    }
    // DLn("after redundant set of kp, ki, kd etc.");
    show_constants();
    EEPROM_writeAnything(0, config_copy);
    for(uint8_t i = 0 ; i < MAX_PROGRAMS ; i++) {
	if(!config_copy.programs[i].name[0]) {
	    final_program = i;
	    break;
	}
    }
#if 0
    // defaults for Cal's house:
    config_copy.offdiff[0] = 1;
    config_copy.ondiff[0]  = 2;
    config_copy.kp      = 45.0;
    config_copy.ki      = 0.0000;
#endif
    DLn("after find final program");
    current_program = config_copy.last_program_used;
    set_temp = config_copy.programs[current_program].temp_min[0].temp;

    lcd.begin(20,4);
    lcd.setBacklightPin(3, POSITIVE);
    lcd.setBacklight(HIGH);
    lcd.clear();
    lcd.print(F(" Snow Farm Creamery "));
    lcd.setCursor(0, 1);
    lcd.print(F("Cheese Heat Control "));
    lcd.setCursor(0, 2);
    lcd.print(F("by William F. Dudley"));
    lcd.setCursor(0, 3);
    lcd.print(F("phone 310 707 8245  "));

    Serial.println("\nArduino Cheese Heater with Webserver");
    DVFn("final_program =", final_program);
    pinMode(BUTTON_0, INPUT); digitalWrite(BUTTON_0, HIGH); // button w/pull-up
    pinMode(BUTTON_1, INPUT); digitalWrite(BUTTON_1, HIGH);
    pinMode(BUTTON_2, INPUT); digitalWrite(BUTTON_2, HIGH);
    pinMode(BUTTON_3, INPUT); digitalWrite(BUTTON_3, HIGH);

    if(!digitalRead(BUTTON_1)) { // did press "2", so use static IP dud's house
	config_copy.ip[2] =  27;
	config_copy.ip[3] =  39;
	config_copy.gateway_ip[2] = config_copy.ip[2];
	goto staticIP;
    }
    else if(!digitalRead(BUTTON_3)) { // did press "4", so use static IP
	uint8_t i;
staticIP:
	DLF("Using static IP: ");
	printlnIP(config_copy.ip);
	DLF("Using gateway: ");
	printlnIP(config_copy.gateway_ip);
	DLF("Using dns: ");
	printlnIP(config_copy.dns_server);
	DLF("Using subnet: ");
	printlnIP(config_copy.subnet_mask);
	Ethernet.begin(config_copy.mac, config_copy.ip, config_copy.dns_server, config_copy.gateway_ip, config_copy.subnet_mask);
	if(!Ethernet.localIP()[0]) {
	    DLFn("Failed to configure Ethernet using static IP");
	    network_failed = 1;
	}
    }
    else {
	DLFn("Using DHCP");
	if(Ethernet.begin(config_copy.mac) == 0) {
	    DLFn("Failed to configure Ethernet using DHCP");
	    network_failed = 1;
	}
	else { // print your local IP address:
	    DLF("IP address:");
#if SEEED_ETHERNET
	    Serial.println(Ethernet.localIP());
#else
	    printlnIP(&Ethernet.localIP()[0]);
#endif
	    // lcd.clear();
	    // lcd.print(F(" Snow Farm Creamery "));
	    // lcd.setCursor(0, 1);
	    // lcd.print(F("Cheese Heat Control "));
	}
    }
    if(network_failed) {
	LCDclearLine (2);
	lcd.print(F("network failed"));
	LCDclearLine (3);
	lcd.print(F("no network"));
    }
    else {
	LCDclearLine (2);
	lcd.print(F("IP address:"));
	LCDclearLine (3);
	LCDprintIP(&Ethernet.localIP()[0]);
    }
    mydelay(3000);
    if(!digitalRead(BUTTON_2)) { // did press "3", so check DNS
	DNSClient dns;
	IPAddress remote_addr;
	int ret;
	DLFn("before dns.begin()");
	dns.begin(Ethernet.dnsServerIP());
	DLFn("before dns.getHostByName('pool.ntp.org')");
	ret = dns.getHostByName("pool.ntp.org", remote_addr);
	// ret = 0;
	if (ret == 1) {
	    DLF("address of pool.ntp.org: ");
	    printlnIP(&remote_addr[0]);
	    dns_working = 1;
	} else {
	    DLFn("(1) dns failed");
//	    lcd.setCursor(0, 1);
//	    LCDserial.print(F("dns failed     ")); SERDELAY;
	    lcd.println(F("dns failed     "));
	    dns_working = 0;
	}
	mydelay(3000);
    }
#define TimerDEBUG 1
#if TimerDEBUG
    Serial.println(F("before setupTimer()"));
#endif
    setupTimer();
#if TimerDEBUG
    Serial.println(F("post setupTimer()"));
#endif

    setUIstate(IDLE);
    sec_Timers[LCD_UPDATE_TIMER] = LCD_UPDATE_PER;
    msec_Timers[BLINK_LCD] = BLINK_MSEC;
    trcount = 0;
}

const char tablestart[] PROGMEM = "<TABLE border=1>\n";

const char programtmpl[] PROGMEM =
"<tr>\n"
"<th colspan=3 align=center>Program</th><td colspan=3 align=center>_NAME &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 1</th><td align=center>_TEMP0 &nbsp;</td>\n"
"<th align=center>Ramp 1</th><td align=center>_RAMP0 &nbsp;</td>\n"
"<th align=center>Time 1</th><td align=center>_TIME0 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 2</th><td align=center>_TEMP1 &nbsp;</td>\n"
"<th align=center>Ramp 2</th><td align=center>_RAMP1 &nbsp;</td>\n"
"<th align=center>Time 2</th><td align=center>_TIME1 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 3</th><td align=center>_TEMP2 &nbsp;</td>\n"
"<th align=center>Ramp 3</th><td align=center>_RAMP2 &nbsp;</td>\n"
"<th align=center>Time 3</th><td align=center>_TIME2 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 4</th><td align=center>_TEMP3 &nbsp;</td>\n"
"<th align=center>Ramp 4</th><td align=center>_RAMP3 &nbsp;</td>\n"
"<th align=center>Time 4</th><td align=center>_TIME3 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 5</th><td align=center>_TEMP4 &nbsp;</td>\n"
"<th align=center>Ramp 5</th><td align=center>_RAMP4 &nbsp;</td>\n"
"<th align=center>Time 5</th><td align=center>_TIME4 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 6</th><td align=center>_TEMP5 &nbsp;</td>\n"
"<th align=center>Ramp 6</th><td align=center>_RAMP5 &nbsp;</td>\n"
"<th align=center>Time 6</th><td align=center>_TIME5 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 7</th><td align=center>_TEMP6 &nbsp;</td>\n"
"<th align=center>Ramp 7</th><td align=center>_RAMP6 &nbsp;</td>\n"
"<th align=center>Time 7</th><td align=center>_TIME6 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 8</th><td align=center>_TEMP7 &nbsp;</td>\n"
"<th align=center>Ramp 8</th><td align=center>_RAMP7 &nbsp;</td>\n"
"<th align=center>Time 8</th><td align=center>_TIME7 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 9</th><td align=center>_TEMP8 &nbsp;</td>\n"
"<th align=center>Ramp 9</th><td align=center>_RAMP8 &nbsp;</td>\n"
"<th align=center>Time 9</th><td align=center>_TIME8 &nbsp;</td>\n"
"</tr>\n"
"<tr>\n"
"<th align=center>Temp 10</th><td align=center>_TEMP9 &nbsp;</td>\n"
"<th align=center>Ramp 10</th><td align=center>_RAMP9 &nbsp;</td>\n"
"<th align=center>Time 10</th><td align=center>_TIME9 &nbsp;</td>\n"
"</tr>\n"
"_EOF\n";

P(root_link) = "<A HREF=\"/\">Back to Top</A><br>";
P(prog_link) = "<A HREF=\"/programs\">Product Programs</A><br>";
P(netw_link) = "<A HREF=\"/network\">Network Settings</A><br>";
P(sett_link) = "<A HREF=\"/settings\">Heat Control Settings</A><br>";
P(log_link)  = "<A HREF=\"/log\">View Run Log</A><br>";
P(graph_link)  = "<A HREF=\"/graph\">View Run Graph</A><br>";
P(angle_br) = "<br>";
P(htmlHead) =
    "<html xmlns=\"http://www.w3.org/1999/xhtml\">"
    "<head>"
    "<title>Snow Farm Creamery Cheese Heat Controller</title>"
    "<style type=\"text/css\">"
    "BODY { font-family: sans-serif }"
    "H1 { font-size: 14pt; text-decoration: underline }"
    "P  { font-size: 10pt; }"
    "</style>"
    "</head>"
    "<body>";

P(htmlTail) =
    "</body>"
    "</html>";

P(formend) = "<input type='submit' value='Submit'></form>";
P(progformstart) = "<form action='" PREFIX "/programs' method='post'>\n";
P(netwformstart) = "<form action='" PREFIX "/network' method='post'>\n";
P(settformstart) = "<form action='" PREFIX "/settings' method='post'>\n";

// read a newline terminated line from the web template
PGM_P read_template_line (char *dbuf, PGM_P s) {
char c;
    *dbuf = '\0';
    while (c = pgm_read_byte(s)) {
	if(c == '\0') {
	    *dbuf++ = '\0';
	    return NULL;
	}      // leave pointing to '\0'
	else if(c == '\n') {
	    *dbuf++ = c;
	    *dbuf++ = '\0';
	    s++;	// point to next line or '\0'
	    break;
	}
	else {
	    *dbuf++ = c;
	    s++;
	}
    }
    return(s);
}

void show_template(WebServer &server, WebServer::ConnectionType type, unsigned char prcnt = 0, PGM_P ptmpl = programtmpl) {
char *p, *d;
unsigned char phase;
char Dbuf3[1000];
    do {
	ptmpl = read_template_line(Dbuf, ptmpl);
	for(p = Dbuf, d = Dbuf3; *p ; ) {
	    if(*p == '_') {
		*d = '\0';
		if(!strncmp_P(p, PSTR("_EOF"), 4)) {
		    ptmpl = NULL;
		    return;
		}
		else if(!strncmp_P(p, PSTR("_NAME"), 5)) {
		    strcat(d, config_copy.programs[prcnt].name);
		    d = Dbuf3 + strlen(Dbuf3);
		    p += 5;
		}
		else if(!strncmp_P(p, PSTR("_TIME"), 5)) {
		    phase = p[5] - '0';
		    sprintf_P(d, PSTR("%d"), config_copy.programs[prcnt].temp_min[phase].minutes);
		    d = Dbuf3 + strlen(Dbuf3);
		    p += 6;
		}
		else if(!strncmp_P(p, PSTR("_TEMP"), 5)) {
		    phase = p[5] - '0';
		    sprintf_P(d, PSTR("%d"), config_copy.programs[prcnt].temp_min[phase].temp);
		    d = Dbuf3 + strlen(Dbuf3);
		    p += 6;
		}
		else if(!strncmp_P(p, PSTR("_RAMP"), 5)) {
		    phase = p[5] - '0';
		    sprintf_P(d, PSTR("%d"), config_copy.programs[prcnt].temp_min[phase].ramp_rate);
		    d = Dbuf3 + strlen(Dbuf3);
		    p += 6;
		}
		else {
		    *d++ = *p++;
		}
	    }
	    else if(*p == '%') {
		if(*++p == 'd') {
		    sprintf(d, "%d", prcnt);
		    d++;
		    *d = '\0';
		    p++;
		}
		else {
		    *d++ = *p++;
		}
	    }
	    else if(*p == '\n') {
		*d++ = *p++;
		*p = '\0';
		*d = '\0';
		break;
	    }
	    else {
		*d++ = *p++;
	    }
	}
#if DEBUG_TMPL
	Serial.print(Dbuf);
	DLFn("^^^___Dbuf");
	Serial.print(Dbuf3);
	DLFn("^^^___Dbuf3");
#endif
	server.print(Dbuf3);
    } while(Dbuf[0] && ptmpl);
}

// show the alarm status/control page
void show_prog_page(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
char sname[20];
#if WEB_OUT_DEBUG
    DVF("show_prog_page, addControls =", addControls);
    DVFn(", UIstate =", UIstate);
#endif
    if(addControls) {
	server.httpSuccess("text/html");
    }
    else {
	server.httpSuccess("text/html","Refresh: 10\r\n");
    }
    server.printP(htmlHead);
    server.printP(tablestart);

    if (addControls && UIstate == IDLE) {
	server.printP(progformstart);
	server.printP(tablestart);
	for(uint8_t pc = 0; pc < MAX_PROGRAMS ; pc++) {
	    uint8_t full;
#if WEB_OUT_DEBUG
	    DVF("P ", pc);
	    DVFn("name ", config_copy.programs[pc].name);
#endif
	    server.printP(PSTR("<tr>\n<th colspan=2 align=center>Program</th><td colspan=2 align=center>"));
	    server.printP(PSTR("<input type=\"text\" size=\"10\" maxlength=\"15\" name=\"pname"));
	    server.print(pc);
	    server.printP(PSTR("\" value=\""));
	    full = config_copy.programs[pc].name[0];
	    if(full) {
		server.print(config_copy.programs[pc].name);
	    }
	    server.printP(PSTR("\"> &nbsp;</td>\n</tr>\n"));
	    for(uint8_t tc = 0 ; tc < MAX_PHASES ; tc++) {
#if WEB_OUT_DEBUG
		DVF("N ", tc);
#endif
		server.printP(PSTR("<tr>\n<th align=center>Temp "));
		server.print(tc + 1);
		server.printP(PSTR("</th><td align=center>"));
		server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\"temp"));
		server.print(pc);
		server.print(tc);
		server.printP(PSTR("\" value=\""));
#if WEB_OUT_DEBUG
		DVF(", temp ", config_copy.programs[pc].temp_min[tc].temp);
#endif
		if(full) {
		    server.print(config_copy.programs[pc].temp_min[tc].temp);
		}
		else {
		    server.print("0");
		}
		server.printP(PSTR("\" &nbsp;</td>\n<th align=center>Ramp "));
		server.print(tc + 1);
		server.printP(PSTR("</th><td align=center>"));
		server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\"ramp"));
		server.print(pc);
		server.print(tc);
		server.printP(PSTR("\" value=\""));
#if WEB_OUT_DEBUG
		DVF(", temp ", config_copy.programs[pc].temp_min[tc].ramp_rate);
#endif
		if(full) {
		    server.print(config_copy.programs[pc].temp_min[tc].ramp_rate);
		}
		else {
		    server.print("0");
		}
		server.printP(PSTR("\" &nbsp;</td>\n<th align=center>Time "));
		server.print(tc + 1);
		server.printP(PSTR("</th><td align=center>"));
		server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"time"));
		server.print(pc);
		server.print(tc);
		server.printP(PSTR("\" value=\""));
#if WEB_OUT_DEBUG
		DVFn(", min ", config_copy.programs[pc].temp_min[tc].minutes);
#endif
		if(full) {
		    server.print(config_copy.programs[pc].temp_min[tc].minutes);
		}
		else {
		    server.print("0");
		}
		server.printP(PSTR("\" &nbsp;</td>\n</tr>\n"));
	    }
	}
	server.println(F("</table>"));
	server.printP(formend);
	server.printP(PSTR("<br>"));
    }
    else {	// don't show the form
	server.printP(tablestart);
	for(uint8_t pc = 0; pc < MAX_PROGRAMS ; pc++) {
	    show_template(server, type, pc, programtmpl);
	}
	server.println(F("</table>"));
    }
    server.printP(root_link);
}

#define WEB_IN_DEBUG 0
#define WEB_OUT_DEBUG 0

void output_prog(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
#if WEB_OUT_DEBUG
    DVF("output_prog, addControls =", addControls);
    DVFn(", UIstate =", UIstate);
#endif
    show_prog_page(server, type, addControls && (UIstate == IDLE));
#if WEB_OUT_DEBUG
    DLFn("end output_prog()");
#endif
}

void prog_Cmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
#if WEB_IN_DEBUG
char sbuf[40];
#endif
    if (type == WebServer::POST) {	// if we have web form results
    bool repeat;
    char name[16], value[20];
    uint8_t pc, tc;
    int phase;
    unsigned char sp;
	do {
	    repeat = server.readPOSTparam(name, 16, value, 20);
#define broken_parser 1
	    if(name[0] == 't' && name[1] == 'i') {	// set time (time11, time12, time13 . . . time95)
		pc = name[4] - '0';	// 0 to 9
		tc = name[5] - '0';	// 0 to 9
		int val = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("time[%d] = '%s'"), pc + 1, value);
		Serial.println(Dbuf);
		sprintf_P(Dbuf, PSTR("old = %d, new = %d"), (int)config_copy.programs[pc].temp_min[tc].minutes, (int)val);
		Serial.println(Dbuf);
#endif
		config_copy.programs[pc].temp_min[tc].minutes = val;
	    }
	    else if(name[0] == 't' && name[1] == 'e') {	// set temp (temp11, temp12, temp13 . . . temp95)
		pc = name[4] - '0';	// 0 to 9
		tc = name[5] - '0';	// 0 to 9
		int val = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("temp[%d] = '%s'"), pc + 1, value);
		Serial.println(Dbuf);
		sprintf_P(Dbuf, PSTR("old = %d, new = %d"), (int)config_copy.programs[pc].temp_min[tc].temp, (int)val);
		Serial.println(Dbuf);
#endif
		config_copy.programs[pc].temp_min[tc].temp = val;
	    }
	    else if(name[0] == 'r' && name[1] == 'a') {	// set ramp_rate (ramp11, ramp12, ramp13 . . . ramp95)
		pc = name[4] - '0';	// 0 to 9
		tc = name[5] - '0';	// 0 to 9
		int val = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("ramp[%d] = '%s'"), pc + 1, value);
		Serial.println(Dbuf);
		sprintf_P(Dbuf, PSTR("old = %d, new = %d"), (int)config_copy.programs[pc].temp_min[tc].ramp_rate, (int)val);
		Serial.println(Dbuf);
#endif
		config_copy.programs[pc].temp_min[tc].ramp_rate = val;
	    }
	    else if(name[0] == 'p' && name[1] == 'n') {	// set pname (pname0, pname2, pname3 . . . pname9)
		pc = name[5] - '0';	// 0 to 9
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("pname[%d] = '%s'"), pc + 1, value);
		Serial.println(Dbuf);
		sprintf_P(Dbuf, PSTR("old = %s, new = %s"), (int)config_copy.programs[pc].name, value);
		Serial.println(Dbuf);
#endif
		strncpy(config_copy.programs[pc].name, value, MAX_NAME);
	    }
	    else if(name[0] == 's' && name[1] == 'c') {	// save config (scfg.) (?SAVE)
		pc = name[4] - '0';
		int val = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		DVF("scfg: name", name);
		DVFn(", pc", pc);
		sprintf_P(Dbuf, PSTR("scfg[%d] = '%s' = %d"), pc, value, val);
		Serial.println(Dbuf);
#endif
		if(val == 1) {
		    eeprom_save_state();
		}
	    }
#if !broken_parser
	    else if (strstr_P(Dbuf2,PSTR("GET /reset")) != 0){ // remote reset, just for the heck of it
		resetFunc();
	    }
#endif
	} while (repeat);
	eeprom_save_state();
	server.httpSeeOther(PREFIX "/programs");
    }
    else {
	output_prog(server, type, true);
#if WEB_OUT_DEBUG
	DLFn("progCmd, addControls is true");
#endif
    }
}

void output_netw(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
int i, h;
char sname[20], ename[4];
    if(addControls) {
	server.httpSuccess("text/html");
    }
    else {
	server.httpSuccess("text/html","Refresh: 10\r\n");
    }
    server.printP(htmlHead);

    if (addControls && UIstate == IDLE) {
	server.printP(netwformstart);
    }

    P(network_pagestart) =
	"<TABLE border=1>"
	"<TR><TH colspan=2>Network Configuration</TH></TR>\n";
    server.printP(network_pagestart);
    server.printP(PSTR("<TR><TD>MAC address:</TD>\n<TD>"));
    for(i = 0 ; i < 6 ; i++) {
	if (addControls && UIstate == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 'm';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server.print(config_copy.mac[i], HEX);
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server.print(config_copy.mac[i], HEX);
	    if(i < 5) { server << ' '; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>my IP address:</TD>\n<TD>"));
    for(i = 0 ; i < 4 ; i++) {
	if (addControls && UIstate == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 'i';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server << config_copy.ip[i];
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server << config_copy.ip[i];
	    if(i < 3) { server << '.'; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>gateway IP address:</TD>\n<TD>"));
    for(i = 0 ; i < 4 ; i++) {
	if (addControls && UIstate == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 'g';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server << config_copy.gateway_ip[i];
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server << config_copy.gateway_ip[i];
	    if(i < 3) { server << '.'; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>dns IP address:</TD>\n<TD>"));
    for(i = 0 ; i < 4 ; i++) {
	if (addControls && UIstate == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 'd';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server << config_copy.dns_server[i];
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server << config_copy.dns_server[i];
	    if(i < 3) { server << '.'; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>subnet mask:</TD>\n<TD>"));
    for(i = 0 ; i < 4 ; i++) {
	if (addControls && UIstate == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 's';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server << config_copy.subnet_mask[i];
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server << config_copy.subnet_mask[i];
	    if(i < 3) { server << '.'; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("</table><br>"));

    if (addControls && UIstate == IDLE) {
	server.printP(formend);
	server.printP(PSTR("<br>"));
    }
    server.printP(root_link);

    server.printP(htmlTail);
#if WEB_OUT_DEBUG
    DLFn("end output_netw()");
#endif
}

// handle network config web form
void netw_Cmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
#if WEB_IN_DEBUG
char sbuf[30];
#endif
    if (type == WebServer::POST) {
    bool repeat, change=false;
    char name[11], value[30];
	do {
	    repeat = server.readPOSTparam(name, 11, value, 30);
	    if(name[0] == 'm'
		 && name[1] >= '0'
		 && name[1] <= '5') {	// mac address m0,m1,m2,m3,m4,m5
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int mindex = strtoul(name + 1, NULL, 10);
		int mvalue = strtoul(value, NULL, 16);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), mindex, mvalue);
		Serial.println(Dbuf);
#endif
		if(mvalue < 256) {
		    if(config_copy.mac[mindex] != mvalue) {
			config_copy.mac[mindex] = mvalue;
			change = true;
		    }
		}
	    }
	    else if(name[0] == 'i'
		 && name[1] >= '0'
		 && name[1] <= '3') {	// ip address i0,i1,i2,i3
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int iindex = strtoul(name + 1, NULL, 10);
		int ivalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), iindex, ivalue);
		Serial.println(Dbuf);
#endif
		if(ivalue < 256) {
		    if(config_copy.ip[iindex] != ivalue) {
			config_copy.ip[iindex] = ivalue;
			change = true;
		    }
		}
	    }
	    else if(name[0] == 'd'
		 && name[1] >= '0'
		 && name[1] <= '3') {	// dns address d0,d1,d2,d3
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int iindex = strtoul(name + 1, NULL, 10);
		int ivalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), iindex, ivalue);
		Serial.println(Dbuf);
#endif
		if(ivalue < 256) {
		    if(config_copy.dns_server[iindex] != ivalue) {
			config_copy.dns_server[iindex] = ivalue;
			change = true;
		    }
		}
	    }
	    else if(name[0] == 'g'
		 && name[1] >= '0'
		 && name[1] <= '3') {	// gateway address g0,g1,g2,g3
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int iindex = strtoul(name + 1, NULL, 10);
		int ivalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), iindex, ivalue);
		Serial.println(Dbuf);
#endif
		if(ivalue < 256) {
		    if(config_copy.gateway_ip[iindex] != ivalue) {
			config_copy.gateway_ip[iindex] = ivalue;
			change = true;
		    }
		}
	    }
	    else if(name[0] == 's'
		 && name[1] >= '0'
		 && name[1] <= '3') {	// subnet mask s0,s1,s2,s3
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int iindex = strtoul(name + 1, NULL, 10);
		int ivalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), iindex, ivalue);
		Serial.println(Dbuf);
#endif
		if(ivalue < 256) {
		    if(config_copy.subnet_mask[iindex] != ivalue) {
			config_copy.subnet_mask[iindex] = ivalue;
			change = true;
		    }
		}
	    }
	} while (repeat);
	server.httpSeeOther(PREFIX "/network");
	if(change) {
	    EEPROM_writeAnything(0, config_copy);
	}
    }
    else {
	output_netw(server, type, true);
    }
}

void output_sett(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
int i, h;
char sname[20], ename[4];
    if(addControls) {
	server.httpSuccess("text/html");
    }
    else {
	server.httpSuccess("text/html","Refresh: 10\r\n");
    }
    server.printP(htmlHead);

    if (addControls && UIstate == IDLE) {
	server.printP(settformstart);
    }

    P(settings_pagestart) =
	"<TABLE border=1>"
	"<TR><TH colspan=2>Heat Control Settings</TH></TR>\n";
    server.printP(settings_pagestart);
    // --------------------------------------------------------
    server.printP(PSTR("<TR><TD colspan=2 align=\"center\">Temp Calibration</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>add this to sensor reading (&deg;F)</TD><TD>"));
    if (addControls && UIstate == IDLE) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"tf\" value =\""));
	server.print(config_copy.tempfudge);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.tempfudge);
    }
    server.printP(PSTR("</TD></TR>\n"));
    // --------------------------------------------------------
    server.printP(PSTR("<TR><TH colspan=2>Bang-Bang Control Settings</TH></TR>\n"));
    // --------------------------------------------------------
    server.printP(PSTR("<TR><TD colspan=2>difference below setpoint when heat turns</TD></TR>\n"));
    server.printP(PSTR("<TR><TD colspan=2>for temps < 101&deg;F</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>Off:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && !config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"fd\" value =\""));
	server.print(config_copy.offdiff[0]);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.offdiff[0]);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>On:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && !config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"nd\" value =\""));
	server.print(config_copy.ondiff[0]);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.ondiff[0]);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD colspan=2>Off value must always be less than On value!</TD></TR>\n"));
    // --------------------------------------------------------
    server.printP(PSTR("<TR><TD colspan=2>for temps 101&deg;F to 120&deg;F</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>Off:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && !config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"f2\" value =\""));
	server.print(config_copy.offdiff[1]);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.offdiff[1]);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>On:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && !config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"n2\" value =\""));
	server.print(config_copy.ondiff[1]);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.ondiff[1]);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD colspan=2>Off value must always be less than On value!</TD></TR>\n"));
    // --------------------------------------------------------
    server.printP(PSTR("<TR><TD colspan=2>for temps > 120&deg;F</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>Off:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && !config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"f3\" value =\""));
	server.print(config_copy.offdiff[2]);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.offdiff[2]);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>On:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && !config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"n3\" value =\""));
	server.print(config_copy.ondiff[2]);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.ondiff[2]);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD colspan=2>Off value must always be less than On value!</TD></TR>\n"));
    // --------------------------------------------------------
    server.printP(PSTR("<TR><TH colspan=2>PID parameters</TH></TR>\n"));
    server.printP(PSTR("<TR><TD>Use PID control</TD>\n<TD>"));
    if (addControls && UIstate == IDLE) {
	if(config_copy.use_pid) {
	    server.printP(PSTR("<label><input type=\"radio\" name=\"up\" value =\"1\" checked>Yes</label><label><input type=\"radio\" name=\"up\" value =\"0\">No</label>"));
	}
	else {
	    server.printP(PSTR("<label><input type=\"radio\" name=\"up\" value =\"1\">Yes</label><label><input type=\"radio\" name=\"up\" value =\"0\" checked>No</label>"));
	}
    }
    else {
	server.print((config_copy.use_pid) ? "Yes" : "No");
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>KP:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"8\" maxlength=\"8\" name=\"kp\" value =\""));
	server.print(config_copy.kp, 6);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.kp, 6);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>KI:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"8\" maxlength=\"8\" name=\"ki\" value =\""));
	server.print(config_copy.ki, 6);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.ki, 6);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>KD:</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"8\" maxlength=\"8\" name=\"kd\" value =\""));
	server.print(config_copy.kd, 6);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.kd, 6);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>PWM period (sec):</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\"pp\" value =\""));
	server.print(config_copy.pwm_period_sec);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.pwm_period_sec);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>Min PWM (sec):</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"mp\" value =\""));
	server.print(config_copy.min_pwm_sec);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.min_pwm_sec);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>Heater Start Delay (sec):</TD>\n<TD>"));
    if (addControls && UIstate == IDLE && config_copy.use_pid) {
	server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\"hd\" value =\""));
	server.print(config_copy.heater_start_delay);
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.print(config_copy.heater_start_delay);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("</table><br>"));

    if (addControls && UIstate == IDLE) {
	server.printP(formend);
	server.printP(PSTR("<br>"));
    }
    server.printP(root_link);

    server.printP(htmlTail);
#if WEB_OUT_DEBUG
    DLFn("end output_sett()");
#endif
}

void output_log(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
char sname[30];
uint8_t logi;
int safei;
    server.httpSuccess("text/html");
    server.printP(htmlHead);
    server.printP(PSTR("<PRE>\n"));
    safei = 0;
    for(logi = 0 ; logi < config_copy.log_index; logi++) {
#if MAXLOGENTRY < 256
	if(logi >= MAXLOGENTRY) { logi = 0; }
#endif
	strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)run_log[logi].UIstate])), sizeof(sname));
	server.print(sname);
	if(run_log[logi].sec_Timer) {
	    server.print(": Time Left = ");
	    server.print((float)run_log[logi].sec_Timer/60.0);
	}
	else if(run_log[logi].UIstate == RUN || run_log[logi].UIstate == PAUSE) {
	    server.print(": acquiring temp ");
	}
	else {
	    server.print(": not regulating ");
	}
	server.print(": sp = ");
	server.print(run_log[logi].set_temp);
	server.print(": spi = ");
	server.print(run_log[logi].set_temp_i);
	if(!config_copy.use_pid) {
	  server.print(": tr = ");
	  server.print(run_log[logi].temp_reached);
	}
	server.print(": state = ");
	server.print(run_log[logi].UIstate);
	server.print(": temp = ");
	server.print(run_log[logi].tempF);
	server.printP(PSTR("\n"));
	if(++safei >= MAXLOGENTRY) { break; }
    }
    server.printP(PSTR("</PRE>\n"));
    server.printP(root_link);

    server.printP(htmlTail);
}

// handle heat control settings web form
void sett_Cmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
#if WEB_IN_DEBUG
char sbuf[30];
    DLFn("in sett_Cmd()");
#endif
    if (type == WebServer::POST) {
    bool repeat, change=false;
    char name[11], value[30];
	do {
	    repeat = server.readPOSTparam(name, 11, value, 30);
	    if(name[0] == 'f'
	    && name[1] == 'd') {	// fd is offdiff[0]
		int mvalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %d"), name, value, mvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.offdiff[0] != mvalue
		&& mvalue < config_copy.ondiff[0]) {
		    config_copy.offdiff[0] = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'n'
	    && name[1] == 'd') {	// nd is ondiff[0]
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int mvalue = strtoul(value, NULL, 10);
		if(config_copy.ondiff[0] != mvalue
		&& mvalue > config_copy.offdiff[0]) {
		    config_copy.ondiff[0] = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'f'
	    && name[1] == '2') {	// f2 is offdiff[1]
		int mvalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %d"), name, value, mvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.offdiff[1] != mvalue
		&& mvalue < config_copy.ondiff[1]) {
		    config_copy.offdiff[1] = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'n'
	    && name[1] == '2') {	// n2 is ondiff[1]
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int mvalue = strtoul(value, NULL, 10);
		if(config_copy.ondiff[1] != mvalue
		&& mvalue > config_copy.offdiff[1]) {
		    config_copy.ondiff[1] = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'f'
	    && name[1] == '3') {	// f3 is offdiff[2]
		int mvalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %d"), name, value, mvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.offdiff[2] != mvalue
		&& mvalue < config_copy.ondiff[2]) {
		    config_copy.offdiff[2] = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'n'
	    && name[1] == '3') {	// n3 is ondiff[2]
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int mvalue = strtoul(value, NULL, 10);
		if(config_copy.ondiff[2] != mvalue
		&& mvalue > config_copy.offdiff[2]) {
		    config_copy.ondiff[2] = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 't'
	    && name[1] == 'f') {	// tf is tempfudge
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		float mvalue = atof(value);
		if(config_copy.tempfudge != mvalue) {
		    config_copy.tempfudge = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'k'
	    && name[1] == 'p') {	// kp is K Proportional
		float fvalue = atof(value);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %f"), name, value, fvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.kp != fvalue) {
		    config_copy.kp = fvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'k'
	    && name[1] == 'i') {	// ki is K Integral
		float fvalue = atof(value);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %f"), name, value, fvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.ki != fvalue) {
		    config_copy.ki = fvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'k'
	    && name[1] == 'd') {	// kd is K Derivative
		float fvalue = atof(value);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %f"), name, value, fvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.kd != fvalue) {
		    config_copy.kd = fvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'u'
	    && name[1] == 'p') {	// up is use_pid
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int mvalue = strtoul(value, NULL, 10);
		if(config_copy.use_pid != mvalue) {
		    config_copy.use_pid = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'p'
	    && name[1] == 'p') {	// pp is pwm_period_sec
		int mvalue = atoi(value);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %d"), name, value, mvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.pwm_period_sec != mvalue) {
		    config_copy.pwm_period_sec = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'm'
	    && name[1] == 'p') {	// mp is min_pwm_sec
		uint8_t mvalue = (uint8_t)atoi(value);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %d"), name, value, (int)mvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.min_pwm_sec != mvalue) {
		    config_copy.min_pwm_sec = mvalue;
		    change = true;
		}
	    }
	    else if(name[0] == 'h'
	    && name[1] == 'd') {	// hd is heater_start_delay
		uint8_t mvalue = (uint8_t)atoi(value);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s = %d"), name, value, (int)mvalue);
		Serial.println(Dbuf);
#endif
		if(config_copy.heater_start_delay != mvalue) {
		    config_copy.heater_start_delay = mvalue;
		    change = true;
		}
	    }
	} while (repeat);
	server.httpSeeOther(PREFIX "/settings");
	if(change) {
	    EEPROM_writeAnything(0, config_copy);
	}
    }
    else {
	output_sett(server, type, true);
    }
}

void log_Cmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
    output_log(server, type, true);
}

void graph_Cmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
    output_graph(server, type, true);
}

void defaultCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
int i;
char sname[20];
    server.httpSuccess("text/html","Refresh: 10\r\n");
    server.printP(htmlHead);

    server << "<h1>Cheese Heater</h1><p>";
    // show system state
    strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)UIstate])), sizeof(sname));
    server << "<h4>State is: ";
    server << sname;
    server << "</h4>";
    server << "<B>Cheese type is: ";
    server << config_copy.programs[current_program].name;
    server << "</B></br>\n";
    server << "<B>Cheese temperature: ";
    server << Fvalue;
    server << "&deg;F</B></br>\n";
    if(UIstate == RUN || UIstate == PAUSE) {
	server << "<B>Temp Set Point is: ";
	server << set_temp;
	server << "&deg;F</B><br>\n";
	if(set_temp_i) {
	    server << "<B>Rate Limited Temp Set Point is: ";
	    server << set_temp_i;
	    server << "&deg;F</B><br>\n";
	}
	server << "<B>Time remaining is: ";
	server << sec_Timers[HEAT_TIMER] / 60;
	server << " minutes</B><br>\n";
    }
    server << "<br>\n";

    server.printP(prog_link);
    server.printP(netw_link);
    server.printP(sett_link);
    server.printP(log_link);
    server.printP(graph_link);

    server.printP(htmlTail);
}

uint8_t ephase;
uint8_t efield;
uint8_t cpos;
char my_alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ 123456789";
char nstr[2];

unsigned long heat_off_time;
unsigned long heat_wait_end;

void set_nstr (void) {
    nstr[1] = '\0';
    if(cpos && isalpha(eprogram.name[cpos-1])) {
	nstr[0] = tolower(my_alphabet[ephase]);
    }
    else {
	nstr[0] = my_alphabet[ephase];
    }
}

void setUIstate (uint8_t new_state) {
//  DVFn("setUIstate ", new_state);
    last_UIstate = UIstate;
    switch(new_state) {
	case IDLE :
	    msec_Timers[READ_TEMP] = TEMP_MSEC;
	    read_temp_now = 1;
	    sec_Timers[HEAT_TIMER] = 0;
	    digitalWrite(HEAT_RELAY, 1 ^ HEAT_RELAY_POLARITY);	// off
	    tone(8, NOTE_A4, 200);
	    phase = 0;
	    set_temp = config_copy.programs[current_program].temp_min[0].temp;
	    break;
	case RUN :
	    msec_Timers[READ_TEMP] = TEMP_MSEC;
	    if(last_UIstate != PAUSE) {	// if PAUSE, just beep once.
		config_copy.log_index = 0;
		phase = 0;
		phase_ended = 0;
		ramp_ended = 0;
		heat_wait_end = 0L;
		integral = 0.0;
		last_error = 0.0;
		temp_reached = 0;
		phase_time = config_copy.programs[current_program].temp_min[0].minutes;
		Fvalue = getTemp();
		set_temp = config_copy.programs[current_program].temp_min[0].temp;
		show_constants();
		if(0) {
		    DVF("current_program =", current_program);
		    DVF("phase =", phase);
		    DVF("phase_time =", phase_time);
		    DVFn("set_temp =", set_temp);
		}
		ramp_rate = config_copy.programs[current_program].temp_min[0].ramp_rate;
		if(!ramp_rate) {
		    ramp_rate = 100;	// get there as fast as you can
		}
		set_temp_i = Fvalue + ramp_rate;
		if(set_temp_i > set_temp) {
		    set_temp_i = set_temp;
		}
		ramp_ended = 1;	// causes do PID to start HEAT_RAMP_TIMER
		// DVFn("setUIstate(RUN): setting ramp_ended=1 & set_temp_i =", set_temp_i);
		tone(8, NOTE_C5, 125);
		mydelay(250);
	    }
	    tone(8, NOTE_C5, 500);
	    break;
	case PAUSE :
	    tone(8, NOTE_C5, 125);
	    msec_Timers[READ_TEMP] = TEMP_MSEC;
	    break;
	case CHANGE_PROG :
	    msec_Timers[READ_TEMP] = 0;
	    sec_Timers[HEAT_TIMER] = 0;
	    break;
	case EDIT :
	    memcpy((void *)&eprogram.name[0], (void *)&config_copy.programs[current_program].name[0], sizeof(CPROGRAM));
	case BANG :
	case SET1 :
	case PID :
	    ephase = 0;
	    efield = 0;
	    break;
	case NEWP :
	    ephase = 0;
	    cpos   = 0;
	    memset((void *)&eprogram.name[0], '\0', sizeof(CPROGRAM));
	    set_nstr();
	    break;
	case DELP :
	    break;
	case SURE :
	    break;
    }
    UIstate = new_state;
    prev_current_program = 0xff;	// highly unlikely to be valid
    lcd_needs_update;
}

int getTemp(void) {
int fraw, change;
float ftr, ft;
double c;
    c = thermocouple.readFarenheit();
    if(isnan(c)) {
	return Fvalue;
    }
    ftr = (float)c;
    ftr += config_copy.tempfudge;
    // DVFn("raw F =", ftr);
    fraw = round(ftr);
    if(trcount < 16) {
	trcount++;
	if(Fvalue == 0 || ((fraw > 50) && (fraw < 150))) {
	    Ffvalue = ftr;
	    return fraw;
	}
    }
    change = abs(Fvalue = fraw);
    if((change < 10 && Fvalue > 40 && Fvalue < 130)
    || (fraw > 40 && fraw < 130 && (Fvalue <= 40 || Fvalue >= 130))) {
	ft = (ftr / 4.0) - (Ffvalue / 4.0);
	Ffvalue += ft;
	outlier_count = 0;
	return round(Ffvalue);
    }
    else {	// very slow filter for "outlier" values
	if(outlier_count) {
	    if(ftr < 50.0) { ftr = 50.0; }
	    if(ftr > 140.0) { ftr = 140.0; }
	    ft = (ftr / 16.0) - (Ffvalue / 16.0);
	    Ffvalue += ft;
	    return round(Ffvalue);
	}
	else {
	    outlier_count++;
	    DLFn("outlier");
	}
	if(fraw < 50 || Fvalue > 140) {
	    DVF("fraw out of bounds ", fraw);
	    DVF("ftr ", ftr);
	    DVFn("Ffv ", Ffvalue);
	}
    }
    return Fvalue;
}

// if(use_pid), do really slow PWM with period of about 5 minutes,
// and on-time not less than about 10 seconds
// otherwise, bang-bang control

void doPID (void) {
float sp, error, deriv, out;
float ekp, iki, dkd;
    offdiff_i = config_copy.offdiff[0];
    ondiff_i = config_copy.ondiff[0];
    if(set_temp_i > 100) {
	offdiff_i = config_copy.offdiff[1];
	ondiff_i = config_copy.ondiff[1];
    }
    else if(set_temp_i > 110) {
	offdiff_i = config_copy.offdiff[2];
	ondiff_i = config_copy.ondiff[2];
    }
    if(set_temp_i && ramp_ended) {
	int old_set_temp_i;
	old_set_temp_i = set_temp_i;
	DVF("doPID, ramp_ended = 1 & old set_temp_i =", set_temp_i);
	sec_Timers[HEAT_RAMP_TIMER] = PWM_PER_SEC;
	ramp_ended = 0;
	set_temp_i += ramp_rate;
	if(set_temp_i > set_temp) {
	    set_temp_i = set_temp;
	    // DVFn("fixed set_temp_i =", set_temp_i);
	}
	DVFn("new set_temp_i =", set_temp_i);
	if(!heat_on
	// && ramp_rate > 0
	&& set_temp_i > old_set_temp_i
	&& (Fvalue < set_temp_i - offdiff_i)) {
	    DLFn("ramp: turning heat on");
	    digitalWrite(HEAT_RELAY, 0 ^ HEAT_RELAY_POLARITY);  // on
	    heat_on = 1;
	}
    }
    sp = (set_temp_i > 0) ? (float)set_temp_i : (float)set_temp;
    if(config_copy.use_pid) {
	error = sp - Ffvalue;
	if(error < 1.1 && error > -1.1) {
	    DVF("absval PID error =", error);
	    DLFn("< 1, so temp_reached = 1, start HEAT_TIMER");
	    temp_reached = 1;
	    if(!sec_Timers[HEAT_TIMER]) {
		sec_Timers[HEAT_TIMER] = phase_time * 60;
		phase_ended = 0;
		DLFn("starting HEAT_TIMER, phase_ended = 0");
	    }
	}
	if(0 && (seconds & 0x03) == 0) {
	    DL("doPID ");
	    DVF("set_temp ", set_temp);
	    DVF(",set_temp_i ", set_temp_i);
	    DVF(",sp ", sp);
	    DVFn("error ", error);
	}
	deriv = error - last_error;
	integral += error;
	ekp = (float)error * (float)KP;
	out = ((float)error * (float)KP) + ((float)integral * (float)KI) + ((float)deriv * (float)KD);
	if(0 && (seconds & 0x03) == 0) {
	    DVF("KP ", KP);
	    DVF(",KI ", KD);
	    DVFn(",KD ", KD);
	    DVF("out ", out);
	    DVF(" > MIN_PWM_SEC ", MIN_PWM_SEC);
	    DVF(" && millis() ", millis());
	    DVF(" > heat_wait_end ", heat_wait_end);
	    if(out >= MIN_PWM_SEC
	    && millis() > heat_wait_end) {
		DLn(" YES");
	    }
	    else {
		DLn(" NO");
	    }
	}
	if(out > PWM_PER_SEC) { out = PWM_PER_SEC; }	// 100% duty cycle
	if(out >= MIN_PWM_SEC
	&& millis() > heat_wait_end) {
	    digitalWrite(HEAT_RELAY, 0 ^ HEAT_RELAY_POLARITY);	// on
	    out += HEATER_START_DELAY;	// hack for heater start time
	    DVF("Heat on for", out);
	    DLFn("seconds.");
	    heat_off_time = ((long)out * 1000L) + millis();
	    heat_wait_end = (PWM_PER_SEC * 1000L) + millis();
	}
	if(millis() > heat_off_time) {
	    digitalWrite(HEAT_RELAY, 1 ^ HEAT_RELAY_POLARITY);	// off
	    heat_off_time = 0L;
	}
	last_error = error;
    }
    else {
	if(heat_on && (Fvalue >= set_temp_i - offdiff_i)) {
	    DLFn("turning heat off");
	    digitalWrite(HEAT_RELAY, 1 ^ HEAT_RELAY_POLARITY);	// off
	    heat_on = 0;
	}
	if(!heat_on && (Fvalue < set_temp_i - ondiff_i)) {
	    DLFn("turning heat on");
	    digitalWrite(HEAT_RELAY, 0 ^ HEAT_RELAY_POLARITY);  // on
	    heat_on = 1;
	}
	if(!temp_reached && Fvalue >= set_temp - offdiff_i) {
	    DVF("heat was on; Fvalue =", Fvalue);
	    DLFn(", so temp_reached = 1, start HEAT_TIMER");
	    temp_reached = 1;
	    if(ramp_rate > 5) {
		ramp_ended = 1;
		ramp_rate = 0;
	    }
	    if(!sec_Timers[HEAT_TIMER]) {
		sec_Timers[HEAT_TIMER] = phase_time * 60;
		phase_ended = 0;
		DLFn("starting HEAT_TIMER, phase_ended = 0");
	    }
	}
    }
}

void doUI (char k) {
int8_t c_p, found_c_p, l_c_p;
    // turn_on_backlight = 1;
    lcd.setBacklight(HIGH);
    // DVF("doUI ", k);
    // DVFn(", UIstate ", UIstate);
    switch(UIstate) {
	case IDLE :
	    switch(k) {
		case 'A' :
		    setUIstate(RUN);
		    break;
		case 'B' :
		    setUIstate(CHANGE_PROG);
		    break;
		case 'C' :
		    setUIstate(EDIT);
		    break;
		case 'D' :
		    setUIstate(MENU2);
		    break;
	    }
	    break;
	case RUN :
	case PAUSE :
	    switch(k) {
		case 'A' :
		    setUIstate((UIstate == IDLE) ? RUN : IDLE);
		    break;
		case 'B' :
		    setUIstate((UIstate == RUN) ? PAUSE : RUN );
		    break;
		case 'C' :
		    set_temp--;
		    break;
		case 'D' :
		    set_temp++;
		    break;
	    }
	    break;
	case CHANGE_PROG :
	    switch(k) {
		case 'A' :
		    setUIstate(IDLE);
		    config_copy.last_program_used = current_program;
		    eeprom_save_state();
		    break;
		case 'B' :	// delete this program
		    setUIstate(DELP);
		    break;
		case 'C' :	// previous program
		    l_c_p = c_p = current_program;
		    c_p--;
		    if(c_p < 0) { c_p = MAX_PROGRAMS-1; }
		    while(!config_copy.programs[c_p].name[0] && c_p != l_c_p) {
			c_p--;
			if(c_p < 0) { c_p = MAX_PROGRAMS-1; }
		    }
		    goto Set_Prog;
		    break;
		case 'D' :	// next program
		    l_c_p = c_p = current_program;
		    c_p++;
		    if(c_p >= MAX_PROGRAMS) { c_p = 0; }
		    while(!config_copy.programs[c_p].name[0] && c_p != l_c_p) {
			c_p++;
			if(c_p >= MAX_PROGRAMS) { c_p = 0; }
		    }
Set_Prog:	    
		    if(config_copy.programs[c_p].name[0]) {
			current_program = c_p;
		    }
		    phase = 0;
		    set_temp = config_copy.programs[current_program].temp_min[phase].temp;
		    break;
	    }
	    break;
	case EDIT :
	    switch(k) {
		case 'A' :	// next field
		    efield++;
		    if(efield == 3) {
			ephase++;
			efield = 0;
		    }
		    break;
		case 'B' :	// save
		    memcpy((void *)&config_copy.programs[current_program].name[0], (void *)&eprogram.name[0], sizeof(CPROGRAM));
		    eeprom_save_state();
		    setUIstate(IDLE);
		    break;
		case 'C' :	// field value --
		    switch(efield) {
			case 2 :
			    if(eprogram.temp_min[ephase].minutes) {
				eprogram.temp_min[ephase].minutes--;
			    }
			    break;
			case 1 :
			    if(eprogram.temp_min[ephase].ramp_rate) {
				eprogram.temp_min[ephase].ramp_rate--;
			    }
			    break;
			case 0 :
			    if(eprogram.temp_min[ephase].temp > 70) {
				eprogram.temp_min[ephase].temp--;
			    }
			    else {
				eprogram.temp_min[ephase].temp = 0;
			    }
			    break;
		    }
		    break;
		case 'D' :	// field value ++
		    switch(efield) {
			case 2 :
			    eprogram.temp_min[ephase].minutes++;
			    break;
			case 1 :
			    if(eprogram.temp_min[ephase].ramp_rate == 4) {
				eprogram.temp_min[ephase].ramp_rate = 0;
			    }
			    else if(eprogram.temp_min[ephase].ramp_rate < 4) {
				eprogram.temp_min[ephase].ramp_rate++;
			    }
			    break;
			case 0 :
			    if(eprogram.temp_min[ephase].temp == 0) {
				eprogram.temp_min[ephase].temp = 70;
			    }
			    else if(eprogram.temp_min[ephase].temp < 150) {
				eprogram.temp_min[ephase].temp++;
			    }
			    break;
		    }
		    break;
	    }
	    break;
	case MENU2 :
	    switch(k) {
		case 'A' :	// idle
		    setUIstate(IDLE);
		    break;
		case 'B' :	// new program
		    setUIstate(NEWP);
		    break;
		case 'C' :	// edit Settings
		    setUIstate(SET1);
		    break;
		case 'D' :	// edit PID OR Bang-Bang
		    setUIstate((config_copy.use_pid) ? PID : BANG);
		    break;
	    }
	    break;
	case SET1 :	// show use_pid and temp_fudge
	    switch(k) {
		case 'A' :	// toggle field
		    efield ^= 1;
		    break;
		case 'B' :	// save
		    eeprom_save_state();
		    setUIstate(IDLE);
		    break;
		case 'C' :	// field value --
		    switch(efield) {
			case 0 :
			    config_copy.use_pid ^= 1;
			    break;
			case 1 :
			    config_copy.tempfudge -= 1.0;
			    break;
		    }
		    break;
		case 'D' :	// field value ++
		    switch(efield) {
			case 0 :
			    config_copy.use_pid ^= 1;
			    break;
			case 1 :
			    config_copy.tempfudge += 1.0;
			    break;
		    }
		    break;
	    }
	    break;
	case BANG :
	    switch(k) {
		case 'A' :	// next field
		    efield++;
		    if(efield == 2) {
			ephase++;
			if(ephase > 2) {
			    ephase = 0;
			}
			efield = 0;
		    }
		    break;
		case 'B' :	// save
		    // perform sanity check on bang-bang values
		    // ondiff must be greater than offdiff
		    for(ephase = 0 ; ephase < 3 ; ephase++) {
			if(config_copy.ondiff[ephase] <= config_copy.offdiff[ephase]) {
			    config_copy.ondiff[ephase] = config_copy.offdiff[ephase] + 1;
			}
		    }
		    eeprom_save_state();
		    setUIstate(IDLE);
		    break;
		case 'C' :	// field value --
		    switch(efield) {
			case 0 :
			    if(config_copy.ondiff[ephase] > -8) {
				config_copy.ondiff[ephase]--;
			    }
			    break;
			case 1 :
			    if(config_copy.offdiff[ephase] > -9) {
				config_copy.offdiff[ephase]--;
			    }
			    break;
		    }
		    break;
		case 'D' :	// field value ++
		    switch(efield) {
			case 0 :
			    if(config_copy.ondiff[ephase] < 9) {
				config_copy.ondiff[ephase]++;
			    }
			    break;
			case 1 :
			    if(config_copy.offdiff[ephase] < 8) {
				config_copy.offdiff[ephase]++;
			    }
			    break;
		    }
		    break;
	    }
	    break;
	case PID :
	    switch(k) {
		case 'A' :	// next phase
		    ephase++;
		    if(ephase > 2) {
			ephase = 0;
		    }
		    break;
		case 'B' :	// save
		    eeprom_save_state();
		    setUIstate(IDLE);
		    break;
		case 'C' :	// field value --
		    switch(phase) {
			case 0 :
			    config_copy.kp--;
			    break;
			case 1 :
			    config_copy.ki--;
			    break;
			case 2 :
			    config_copy.kd--;
			    break;
		    }
		    break;
		case 'D' :	// field value ++
		    switch(phase) {
			case 0 :
			    config_copy.kp++;
			    break;
			case 1 :
			    config_copy.ki++;
			    break;
			case 2 :
			    config_copy.kd++;
			    break;
		    }
		    break;
	    }
	    break;
	case NEWP :
	    switch(k) {
		case 'A' :	// save this letter
		    eprogram.name[cpos] = nstr[0];
		    if(cpos < MAX_NAME-1) cpos++;
		    break;
		case 'B' :	// save name, go to EDIT mode
		    memcpy((void *)&config_copy.programs[final_program].name[0], (void *)&eprogram.name[0], sizeof(CPROGRAM));
		    current_program = final_program++;
		    setUIstate(EDIT);
		    break;
		case 'C' :	// previous letter
		    if(ephase) { ephase--; }
		    set_nstr();
		    break;
		case 'D' :	// next letter
		    if(ephase < sizeof(my_alphabet)-1) { ephase++; }
		    set_nstr();
		    break;
	    }
	    break;
	case DELP :
	    switch(k) {
		case 'A' :
		    setUIstate(SURE);
		case 'B' :
		case 'C' :
		case 'D' :
		    setUIstate(IDLE);
		    break;
	    }
	    break;
	case SURE :
	    switch(k) {
		case 'A' :
		    setUIstate(IDLE);
		    break;
		case 'B' :
		    memset((void *)&eprogram.name[0], '\0', sizeof(CPROGRAM));
		    memcpy((void *)&config_copy.programs[current_program].name[0], (void *)&eprogram.name[0], sizeof(CPROGRAM));
		    c_p = current_program;
		    found_c_p = 0;
		    if(c_p) { c_p--; }
		    while(c_p
		    && !config_copy.programs[c_p].name[0]) {
			c_p--;
		    }
		    if(config_copy.programs[c_p].name[0]) {
			current_program = c_p;
			found_c_p = 1;
		    }
		    if(!found_c_p) {
			if(c_p < MAX_PROGRAMS) { c_p++; }
			while(c_p < MAX_PROGRAMS
			&& !config_copy.programs[c_p].name[0]) {
			    c_p++;
			}
			if(config_copy.programs[c_p].name[0]) {
			    current_program = c_p;
			    found_c_p = 1;
			}
		    }
		    if(!found_c_p) {
			current_program = 0;
		    }
		    eeprom_save_state();
		case 'C' :
		case 'D' :
		    setUIstate(IDLE);
		    break;
	    }
	    break;
    }
    lcd_needs_update = 1;
}

char scanButtons (void) {
static uint8_t priors[4];
    if(!digitalRead(BUTTON_3) && priors[3]) { priors[3] = 0; return 'D'; }
    if(digitalRead(BUTTON_3) && !priors[3]) { priors[3] = 1; }
    if(!digitalRead(BUTTON_2) && priors[2]) { priors[2] = 0; return 'C'; }
    if(digitalRead(BUTTON_2) && !priors[2]) { priors[2] = 1; }
    if(!digitalRead(BUTTON_1) && priors[1]) { priors[1] = 0; return 'B'; }
    if(digitalRead(BUTTON_1) && !priors[1]) { priors[1] = 1; }
    if(!digitalRead(BUTTON_0) && priors[0]) { priors[0] = 0; return 'A'; }
    if(digitalRead(BUTTON_0) && !priors[0]) { priors[0] = 1; }
    return '\0';
}

void serial_command(char schar) {
uint8_t change;
    change = 0;
    // DVFn("char =", schar);
    if(schar == '\r' || schar == '\0') {
	// do the command
	if(cbuf[0] == 'n' && cbuf[1] == 'd'
	&& cbuf[2] == '=' && cbuf[3] >= '0' && cbuf[3] <= '9') {
	    config_copy.ondiff[0] = atoi(&cbuf[3]);
	    DVFn("ondiff[0] =", config_copy.ondiff[0]);
	    change++;
	}
	else if(cbuf[0] == 'f' && cbuf[1] == 'd'
	&& cbuf[2] == '=' && cbuf[3] >= '0' && cbuf[3] <= '9') {
	    config_copy.offdiff[0] = atoi(&cbuf[3]);
	    DVFn("offdiff[0] =", config_copy.offdiff[0]);
	    change++;
	}
	else if(cbuf[0] == 'n' && cbuf[1] == 'd' && cbuf[2] == '2'
	&& cbuf[3] == '=' && cbuf[4] >= '0' && cbuf[4] <= '9') {
	    config_copy.ondiff[1] = atoi(&cbuf[4]);
	    DVFn("ondiff[1] =", config_copy.ondiff[1]);
	    change++;
	}
	else if(cbuf[0] == 'f' && cbuf[1] == 'd' && cbuf[2] == '2'
	&& cbuf[3] == '=' && cbuf[4] >= '0' && cbuf[4] <= '9') {
	    config_copy.offdiff[1] = atoi(&cbuf[4]);
	    DVFn("offdiff[1] =", config_copy.offdiff[1]);
	    change++;
	}
	else if(cbuf[0] == 'n' && cbuf[1] == 'd' && cbuf[2] == '3'
	&& cbuf[3] == '=' && cbuf[4] >= '0' && cbuf[4] <= '9') {
	    config_copy.ondiff[2] = atoi(&cbuf[4]);
	    DVFn("ondiff[2] =", config_copy.ondiff[2]);
	    change++;
	}
	else if(cbuf[0] == 'f' && cbuf[1] == 'd' && cbuf[2] == '3'
	&& cbuf[3] == '=' && cbuf[4] >= '0' && cbuf[4] <= '9') {
	    config_copy.offdiff[2] = atoi(&cbuf[4]);
	    DVFn("offdiff[2] =", config_copy.offdiff[2]);
	    change++;
	}
	else if(cbuf[0] == 'k' && cbuf[1] == 'p'
	&& cbuf[2] == '=' && cbuf[3] >= '0' && cbuf[3] <= '9') {
	    config_copy.kp = atof(&cbuf[3]);
	    DVFn("KP =", config_copy.kp);
	    change++;
	}
	else if(cbuf[0] == 'k' && cbuf[1] == 'i'
	&& cbuf[2] == '=' && cbuf[3] >= '0' && cbuf[3] <= '9') {
	    config_copy.ki = atof(&cbuf[3]);
	    DVFn("KI =", config_copy.ki);
	    change++;
	}
	else if(cbuf[0] == 'k' && cbuf[1] == 'd'
	&& cbuf[2] == '=' && cbuf[3] >= '0' && cbuf[3] <= '9') {
	    config_copy.kd = atof(&cbuf[3]);
	    DVFn("KD =", config_copy.kd);
	    change++;
	}
	else if(cbuf[0] == 't' && cbuf[1] == 'f'
	&& cbuf[2] == '=' && cbuf[3] >= '0' && cbuf[3] <= '9') {
	    config_copy.tempfudge = atof(&cbuf[3]);
	    DVFn("TempFudge =", config_copy.tempfudge);
	    change++;
	}
	else if(cbuf[0] == 'u' && cbuf[1] == 'p'
	&& cbuf[2] == '=' && cbuf[3] >= '0' && cbuf[3] <= '1') {
	    config_copy.use_pid = atoi(&cbuf[3]);
	    DVFn("use pid =", config_copy.use_pid);
	    change++;
	}
	else if(cbuf[0] == 'L' && cbuf[1] == 'o' && cbuf[2] == 'g') {
	    populate_log();	// just for testing log feature
	}
	else if(cbuf[0] == 'l' && cbuf[1] == 'o' && cbuf[2] == 'g') {
	    char sname[50];
	    int safei;
	    DLFn("-----------------------------------------------------------");
	    safei = 0;
	    for(uint8_t logi = 0 ; logi < config_copy.log_index; logi++) {
#if MAXLOGENTRY < 256
		if(logi >= MAXLOGENTRY) { logi = 0; }
#endif
		strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)run_log[logi].UIstate])), sizeof(sname));
		Serial.print(sname);
		if(run_log[logi].sec_Timer) {
		    DLF(": Time Left = ");
		    Serial.print((float)run_log[logi].sec_Timer/60.0);
		}
		else if(run_log[logi].UIstate == RUN || run_log[logi].UIstate == PAUSE) {
		    DLF(": acquiring temp ");
		}
		else {
		    DLF(": not regulating ");
		}
		DLF(": sp = ");
		Serial.print(run_log[logi].set_temp);
		DLF(": spi = ");
		Serial.print(run_log[logi].set_temp_i);
		if(!config_copy.use_pid) {
		  Serial.print(": tr = ");
		  Serial.print(run_log[logi].temp_reached);
		}
		DLF(": state = ");
		Serial.print(run_log[logi].UIstate);
		DLF(": temp = ");
		Serial.print(run_log[logi].tempF);
		DLF("\n");
		if(++safei >= MAXLOGENTRY) { break; }
	    }
	    DLFn("-----------------------------------------------------------");
	}
	else if(cbuf[0] == '?') {
	    show_constants();
	}
	else if(cbuf[0] == 'S' && cbuf[1] == 'a'
	     && cbuf[2] == 'v' && cbuf[3] == 'e') {
	    DLF("Saving . . .");
	    save_log_to_eeprom ();
	    eeprom_save_state();
	    DLFn(" Done");
	}
	else if(cbuf[0] == 'h') {	// help
	    DLFn("nd=2 set ondiff[0] to 2 (for temps < 101)");
	    DLFn("fd=3 set offdiff[0] to 3 (for temps < 101)");
	    DLFn("nd2=1 set ondiff[1] to 1 (for temps 101-120)");
	    DLFn("fd2=2 set offdiff[1] to 2 (for temps 101-120)");
	    DLFn("nd3=0 set ondiff[2] to 0 (for temps > 120)");
	    DLFn("fd3=1 set offdiff[2] to 1 (for temps > 120)");
	    DLFn("kp=30.0 set KP to 30.0");
	    DLFn("ki=3.0 set KI to 3.0");
	    DLFn("kd=0.002 set KD to 0.002");
	    DLFn("tf=20.1 set Temp Fudge to 20.1");
	    DLFn("up=0 use Bang Bang control");
	    DLFn("up=1 use PID control");
	    DLFn("? show all these variable values");
	    DLFn("no spaces allowed");
	}
	// clear the buffer
	memset(cbuf, '\0', sizeof(cbuf)/sizeof(char));
	ci = 0;
    }
    else {
	cbuf[ci] = schar;
	if(ci < 79) { ci++; }
    }
    if(change) {
	eeprom_save_state();
    }
}

void loop () {
char keypress, schar;
    keypress = scanButtons();
    if(keypress) {
//	DVFn("key =", keypress);
	doUI(keypress);
    }
    if (Serial.available() > 0) {
        schar = Serial.read();
        // DVFn("char =", schar);
        serial_command(schar);
    }
    if(read_temp_now) {
	read_temp_now = 0;
	Fvalue = getTemp();
	if(UIstate == RUN
	|| UIstate == PAUSE) {
	    doPID();
	}
    }

    if(lcd_needs_update) {
	lcd_needs_update = 0;
	if(UIstate != NEWP
	&& prev_current_program != current_program) {
	    lcd.setCursor(0, 0);	// first line
	    lcd.print(F("Prog: "));	// 6 chars
	    if(config_copy.programs[current_program].name[0]) {
		lcd.print(config_copy.programs[current_program].name);
		for(uint8_t len = (uint8_t)strlen(config_copy.programs[current_program].name) ;
		    len <= 20 ; len++) {
		    lcd.print(F(" "));
		}
	    }
	    else {
		lcd.print(F("empty slot    "));
	    }
	    prev_current_program = current_program;
	}
	switch(UIstate) {
	    case IDLE :
	    case PAUSE :
	    case RUN :
		if(set_temp_i) {
		    lcd.setCursor(0, 1);	// second line
		    lcd.print(F("Set "));	// 4 chars
		    lcd.print(set_temp);	// 6 or 7 chars total
		    if(set_temp < 100) {
			lcd.print(F(" "));	// 7 chars total
		    }
		    lcd.print(F("L "));	// 9 chars total
		    lcd.print(set_temp_i);	// 11 or 12 chars total
		    if(set_temp_i < 100) {
			lcd.print(F(" "));	// 12 chars total
		    }
		    lcd.print(F("Now "));	// 16 chars total
		    lcd.print(Fvalue);	// 18 or 19 chars total
		    lcd.print(F("F"));	// 19 or 20 chars total
		    if(Fvalue < 100) {
			lcd.print(F(" "));	// 20 chars total
		    }
		}
		else {
		    lcd.setCursor(0, 1);	// second line
		    lcd.print(F("Set  "));
		    lcd.print(set_temp);
		    lcd.print(F("F    "));
		    lcd.setCursor(10, 1);	// second line, second half
		    lcd.print(F("Act  "));
		    lcd.print(Fvalue);
		    lcd.print(F("F  "));
		}
		break;
	    case CHANGE_PROG :
		lcd.setCursor(0, 1);
		lcd.print(F("Change program mode"));
		break;
	    case EDIT :
		// lcd_needs_update = 1; // because blink needs frequent upd
		lcd.setCursor(0, 1);	// second line
		lcd.print(F("ph "));
		lcd.print(ephase);		// 4 chars
		lcd.print(F(" T "));	// 7 chars
		if(efield == 0 && !blink_state) {
		    lcd.print(F("   "));
		}
		else {
		    lcd.print(eprogram.temp_min[ephase].temp);
		    if(eprogram.temp_min[ephase].temp < 100) {
			lcd.print(F(" "));
		    }
		}
		// 9 or 10 chars
		lcd.print(F(" R "));
		// 12 or 13 chars
		if(efield == 1 && !blink_state) {
		    lcd.print(F(" "));
		}
		else {
		    lcd.print(eprogram.temp_min[ephase].ramp_rate);
		}
		// 14 or 15chars
		lcd.print(F(" D "));
		// 17 or 18 chars
		if(efield == 2 && !blink_state) {
		    lcd.print(F("  "));
		    lcd.print((eprogram.temp_min[ephase].minutes < 100) ? F("m") : F(" "));
		}
		else {
		    lcd.print(eprogram.temp_min[ephase].minutes);
		    // 19 or 20 chars
		    if(eprogram.temp_min[ephase].minutes < 100) {
			lcd.print(F("m"));
		    }
		}
		// 20 chars on line 2

		lcd.setCursor(0, 2);	// third line
		lcd.print(F("EDIT  1=next  2=save"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("Temp  3=down  4=up  "));
		break;
	    case NEWP :
		lcd.clear();
		lcd.print(F("New Program - name: "));
		lcd.setCursor(0, 1);	// second line
		lcd.print(eprogram.name);
		if(!blink_state) {
		    lcd.print(nstr);
		}
		else {
		    lcd.print(F("_"));
		}
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("NEW   1=next  2=save"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("Letter 3=down  4=up "));
		break;
	    case DELP :
		lcd.clear();
		lcd.print(F("Delete Program      "));
		lcd.setCursor(0, 1);	// second line
		lcd.print(config_copy.programs[current_program].name);
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("To delete this prog."));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("press 1; 2 to abort "));
		break;
	    case SURE :
		lcd.clear();
		lcd.print(F("Delete Program      "));
		lcd.setCursor(0, 1);	// second line
		lcd.print(config_copy.programs[current_program].name);
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("If you're certain,  "));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("press 2; 1 to abort "));
		break;
	    case SET1 :
		lcd.clear();
		lcd.print(F("Edit General Setting"));
		lcd.setCursor(0, 1);	// second line
		lcd.print(F("PID "));	// 4 chars
		if(efield == 0 && !blink_state) {
		    lcd.print(F(" "));
		}
		else {
		    lcd.print(config_copy.use_pid); // 5 chars
		}
		lcd.print(F(" tcal "));	// 11 chars
		if(efield == 1 && !blink_state) {
		    lcd.print(F("         "));
		}
		else {
		    lcd.print(config_copy.tempfudge); // ? chars
		}
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("EDIT  1=next  2=save"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("Diff  3=down  4=up  "));
		break;
	    case BANG :
		lcd.clear();
		lcd.print(F("Edit BangBang params"));
		lcd.setCursor(0, 1);	// second line
		switch(ephase) {
		    case 0 :
			lcd.print(F("< 101 on "));	// 9 chars
			break;
		    case 1 :
			lcd.print(F("101-110 on "));	// 11 chars
			break;
		    case 2 :
			lcd.print(F("> 110 on "));	// 9 chars
			break;
		    }
		if(efield == 0 && !blink_state) {
		    lcd.print(F(" "));
		    if(config_copy.ondiff[ephase] < 0) {
			lcd.print(F(" "));
		    }
		}
		else {
		    lcd.print(config_copy.ondiff[ephase]); // 10 or 12 chars
		}
		lcd.print(F(" off "));	// 15 or 17 chars
		if(efield == 1 && !blink_state) {
		    lcd.print(F(" "));
		    if(config_copy.offdiff[ephase] < 0) {
			lcd.print(F(" "));
		    }
		}
		else {
		    lcd.print(config_copy.offdiff[ephase]); // 16 or 18 chars
		}
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("EDIT  1=next  2=save"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("Diff  3=down  4=up  "));
		break;
	    case PID:
		lcd.clear();
		lcd.print(F("Edit PID params     "));
		lcd.setCursor(0, 1);	// second line
		switch(ephase) {
		    case 0 :
			lcd.print(F("KP= "));	// 4 chars
			lcd.print(config_copy.kp);  // 16 chars incl decimal
			break;
		    case 1 :
			lcd.print(F("KI= "));	// 4 chars
			lcd.print(config_copy.ki);	// 16 chars incl decimal
			break;
		    case 2 :
			lcd.print(F("KD= "));	// 4 chars
			lcd.print(config_copy.kd);	// 16 chars incl decimal
			break;
		}
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("EDIT  1=next  2=save"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("pid  3=down  4=up  "));
		break;
	}
	switch(UIstate) {
	    case IDLE :
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("IDLE 1=run 2=program"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("     3=edit 4=more  "));
		break;
	    case RUN :
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("1=stop 2=pause "));
		lcd.print((30 + sec_Timers[HEAT_TIMER]) / 60);
		if(sec_Timers[HEAT_TIMER] < 600) {
		    lcd.print(F(" "));
		}
		lcd.print(F("min"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("Temp  3=down  4=up  "));
		break;
	    case PAUSE :
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("Paused  2=run   "));
		lcd.print((30 + sec_Timers[HEAT_TIMER]) / 60);
		lcd.print(F("m "));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("Temp  3=down  4=up  "));
		break;
	    case CHANGE_PROG :
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("1 = idle, 2 = delete"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("3 = prev, 4 = next  "));
		break;
	    case MENU2 :
		lcd.setCursor(0, 0);	// first line
		lcd.print(F("More Edit "));
		lcd.setCursor(0, 1);	// second line
		lcd.print(F("1 = idle, 2 = newp  "));
		lcd.setCursor(0, 2);	// third line
		lcd.print(F("3 = general settings"));
		lcd.setCursor(0, 3);	// fourth line
		lcd.print(F("4 = edit ctrl vars  "));
		break;
	}
    }
    if(UIstate == RUN) {
	if(phase_ended) {
	    DLFn("H_T expired, RUN mode");
	    digitalWrite(HEAT_RELAY, 1 ^ HEAT_RELAY_POLARITY);	// off
	    tone(8, NOTE_C5, 250);
	    phase++;
	    ramp_ended = 0;
	    set_temp = config_copy.programs[current_program].temp_min[phase].temp;
	    if(set_temp) {	// it's a valid phase
		set_temp_i = 0;
		ramp_rate = config_copy.programs[current_program].temp_min[phase].ramp_rate;
		if(!ramp_rate) {
		    ramp_rate = 100;	// get there as fast as you can
		}
		set_temp_i = Fvalue;
		ramp_ended = 1;	// causes do PID to start HEAT_RAMP_TIMER
		DVFn("loop(): phase_ended, setting ramp_ended=1 & set_temp_i =", set_temp_i);
		phase_time = config_copy.programs[current_program].temp_min[phase].minutes;
		phase_ended = 0;
		integral = 0.0;		// PID
		last_error = 0.0;	// PID
		temp_reached = 0;	// BangBang
		lcd_needs_update++;
	    }
	    else {
		setUIstate(IDLE);
		finished_run = 5;	// keep logging for 5 minutes
		save_log_to_eeprom();
		eeprom_save_state();
	    }
	}
    }
    if(new_minute) {
    char sname[20];
	strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)UIstate])), sizeof(sname));
	DL(sname);
	if(sec_Timers[HEAT_TIMER]) {
	    DVF(": Time Left =", (float)sec_Timers[HEAT_TIMER]/60.0);
	}
	else if(UIstate == RUN || UIstate == PAUSE) {
	    DLF(": acquiring temp ");
	}
	else {
	    DLF(": not regulating ");
	}
	DVF(": sp =", set_temp);
	DVF(": spi =", set_temp_i);
	if(!config_copy.use_pid) {
	  DVF(": tr =", temp_reached);
	}
	DVF(": state =", UIstate);
	DVFn(": temp =", Ffvalue);
	new_minute = 0;
	if(UIstate == RUN || UIstate == PAUSE || finished_run) {
	    store_log(Fvalue);
	    if(finished_run) {
		finished_run--;
		if(!finished_run) {
		    save_log_to_eeprom();
		    eeprom_save_state();
		}
	    }
	}
    }
    if(1) {
	WebServer webserver(PREFIX, 80);
	webserver.begin();

	webserver.setDefaultCommand(&defaultCmd);
	webserver.addCommand("programs", &prog_Cmd);
	webserver.addCommand("network", &netw_Cmd);
	webserver.addCommand("settings", &sett_Cmd);
	webserver.addCommand("log", &log_Cmd);
	webserver.addCommand("graph", &graph_Cmd);

	webserver.processConnection();
    }
}

void save_log_to_eeprom (void) {
    EEPROM_writeAnything(sizeof(config_copy)/sizeof(char), run_log);
}

void restore_log_from_eeprom (void) {
    EEPROM_readAnything(sizeof(config_copy)/sizeof(char), run_log);
}

void populate_log (void) {
uint8_t i;
    i = 0;
    run_log[i++] = (LOGENTRY){ 92, 0, 0, 0, 79, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 79, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 79, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 81, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 82, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 81, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 82, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 83, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 84, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 85, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 85, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 86, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 88, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 89, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 0, 1, 90, 0 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 69 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 90, 68 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 67 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 90, 66 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 90, 65 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 64 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 63 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 62 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 61 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 60 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 59 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 58 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 57 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 56 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 55 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 54 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 53 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 52 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 51 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 50 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 49 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 48 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 47 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 46 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 45 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 44 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 43 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 42 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 41 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 40 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 39 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 38 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 37 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 36 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 35 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 34 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 33 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 32 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 31 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 30 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 29 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 28 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 27 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 26 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 25 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 24 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 23 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 22 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 21 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 20 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 19 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 18 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 17 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 16 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 15 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 14 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 13 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 12 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 11 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 10 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 9 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 8 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 7 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 6 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 5 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 92, 4 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 91, 3 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 93, 2 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 93, 1 * 60 };
    run_log[i++] = (LOGENTRY){ 92, 92, 1, 1, 93, 0 };
    run_log[i++] = (LOGENTRY){ 104, 95, 0, 1, 93, 0 };
    run_log[i++] = (LOGENTRY){ 104, 95, 0, 1, 93, 0 };
    run_log[i++] = (LOGENTRY){ 104, 95, 0, 1, 93, 0 };
    run_log[i++] = (LOGENTRY){ 104, 95, 0, 1, 93, 0 };
    run_log[i++] = (LOGENTRY){ 104, 95, 0, 1, 92, 0 };
    run_log[i++] = (LOGENTRY){ 104, 97, 0, 1, 92, 0 };
    run_log[i++] = (LOGENTRY){ 104, 97, 0, 1, 92, 0 };
    run_log[i++] = (LOGENTRY){ 104, 97, 0, 1, 93, 0 };
    run_log[i++] = (LOGENTRY){ 104, 97, 0, 1, 93, 0 };
    run_log[i++] = (LOGENTRY){ 104, 97, 0, 1, 94, 0 };
    run_log[i++] = (LOGENTRY){ 104, 99, 0, 1, 94, 0 };
    run_log[i++] = (LOGENTRY){ 104, 99, 0, 1, 94, 0 };
    run_log[i++] = (LOGENTRY){ 104, 99, 0, 1, 95, 0 };
    run_log[i++] = (LOGENTRY){ 104, 99, 0, 1, 95, 0 };
    run_log[i++] = (LOGENTRY){ 104, 99, 0, 1, 96, 0 };
    run_log[i++] = (LOGENTRY){ 104, 101, 0, 1, 97, 0 };
    run_log[i++] = (LOGENTRY){ 104, 101, 0, 1, 97, 0 };
    run_log[i++] = (LOGENTRY){ 104, 101, 0, 1, 98, 0 };
    run_log[i++] = (LOGENTRY){ 104, 101, 0, 1, 98, 0 };
    run_log[i++] = (LOGENTRY){ 104, 101, 0, 1, 99, 0 };
    run_log[i++] = (LOGENTRY){ 104, 103, 0, 1, 100, 0 };
    run_log[i++] = (LOGENTRY){ 104, 103, 0, 1, 100, 0 };
    run_log[i++] = (LOGENTRY){ 104, 103, 0, 1, 101, 0 };
    run_log[i++] = (LOGENTRY){ 104, 103, 0, 1, 102, 0 };
    run_log[i++] = (LOGENTRY){ 104, 103, 0, 1, 102, 0 };
    run_log[i++] = (LOGENTRY){ 104, 104, 0, 1, 103, 0 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 103, 19 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 103, 18 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 103, 17 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 104, 16 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 104, 15 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 104, 14 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 104, 13 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 104, 12 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 11 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 10 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 9 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 8 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 7 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 6 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 5 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 4 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 3 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 2 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 1 * 60 };
    run_log[i++] = (LOGENTRY){ 104, 104, 1, 1, 105, 0 };
    run_log[i++] = (LOGENTRY){ 120, 109, 0, 1, 105, 0 };
    run_log[i++] = (LOGENTRY){ 120, 109, 0, 1, 105, 0 };
    run_log[i++] = (LOGENTRY){ 120, 109, 0, 1, 105, 0 };
    run_log[i++] = (LOGENTRY){ 120, 109, 0, 1, 106, 0 };
    run_log[i++] = (LOGENTRY){ 120, 109, 0, 1, 106, 0 };
    run_log[i++] = (LOGENTRY){ 120, 113, 0, 1, 107, 0 };
    run_log[i++] = (LOGENTRY){ 120, 113, 0, 1, 107, 0 };
    run_log[i++] = (LOGENTRY){ 120, 113, 0, 1, 107, 0 };
    run_log[i++] = (LOGENTRY){ 120, 113, 0, 1, 108, 0 };
    run_log[i++] = (LOGENTRY){ 120, 113, 0, 1, 109, 0 };
    run_log[i++] = (LOGENTRY){ 120, 117, 0, 1, 109, 0 };
    run_log[i++] = (LOGENTRY){ 120, 117, 0, 1, 110, 0 };
    run_log[i++] = (LOGENTRY){ 120, 117, 0, 1, 109, 0 };
    run_log[i++] = (LOGENTRY){ 120, 117, 0, 1, 109, 0 };
    run_log[i++] = (LOGENTRY){ 120, 117, 0, 1, 110, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 110, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 111, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 112, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 112, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 113, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 113, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 113, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 114, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 114, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 115, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 115, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 116, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 116, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 116, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 117, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 117, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 118, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 0, 1, 118, 0 };
    run_log[i++] = (LOGENTRY){ 120, 120, 1, 1, 119, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 119, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 119, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 119, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 119, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 119, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 121, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 121, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 121, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 121, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 121, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 121, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 121, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 120, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 119, 0 };
    run_log[i++] = (LOGENTRY){ 92, 120, 1, 0, 119, 0 };
    config_copy.log_index = i;
}

#define CSTMP  "00,00,200"
#define CSTMPI "0,200,0"
#define CACTMP "200,00,00"
char * graph2color[] = { CSTMP, CSTMPI, CACTMP };
int highest_temp, lowest_temp, x, xleft, hm, xstep, yoff;
int vm;		// vertical multiplier
int gmargin, ytop, graphwidth, xmax, vsize;

void scan_log (void) {
uint8_t logi;
int safei;
    DLFn("scan_log");
    lowest_temp  =  200;
    highest_temp =    0;
    graphwidth = 0;
    DVFn("log_index =", config_copy.log_index);
    safei = 0;
    for(logi = 0 ; logi < config_copy.log_index; logi++) {
	// DVF("logi =", logi);
	// if((graphwidth % 5) == 0) { DLF("\n"); }
#if MAXLOGENTRY < 256
	if(logi >= MAXLOGENTRY) { logi = 0; }
#endif
	if(run_log[logi].tempF > 10
	&& run_log[logi].tempF < lowest_temp) {
	    lowest_temp = run_log[logi].tempF;
	}
	if(run_log[logi].tempF > highest_temp) {
	    highest_temp = run_log[logi].tempF;
	}
	graphwidth++;
	if(++safei >= MAXLOGENTRY) { break; }
    }
    DVFn("highest_temp =", highest_temp);
    DVFn("lowest_temp =", lowest_temp);
    DVFn("graphwidth =", graphwidth);
}

#define fontfam = "font-family=\"Verdana\""

int y_lines (WebServer &server, int ymax, int x1, int x2, int ysize, int vmin, int vmax) {
char sbuf[150];
// y1 = yoff + (ytop - (lastrow[graph] - lowest_temp)) * vm
// y2 = yoff + (ytop - (thisrow[graph] - lowest_temp)) * vm
    int lowv = 10 * ((vmin + 5) / 10);
    int highv = 10 * ((vmax / 10));
    // fprintf(stderr, "vmax = %d, highv = %d\n", vmax, highv);
    // if(lowv < vmin) { lowv += 10; }
    // fprintf(stderr, "vmin = %d, int(vmin/10) = %d, lowv = %d, highv = %d\n", vmin, int(vmin / 10), lowv, highv);
    for(int i = lowv ; i <= highv ; i += 2) {
	if((i % 10) == 0) { continue; }
	sprintf_P(sbuf, PSTR("<line x1=\"%d\" y1=\"%d\" x2=\"%d\" y2=\"%d\" style=\"stroke:rgb(10,10,10);stroke-width:1\"/>\n"), x, yoff + (ytop - (i - vmin)) * vm , x2, yoff + (ytop - (i - vmin)) * vm);
	server.print(sbuf);
    }
    for(int i = lowv ; i <= highv ; i += 10) {
	sprintf_P(sbuf, PSTR("<line x1=\"%d\" y1=\"%d\" x2=\"%d\" y2=\"%d\" style=\"stroke:rgb(30,30,30);stroke-width:2\"/>\n"), x, yoff + (ytop - (i - vmin)) * vm , x2, yoff + (ytop - (i - vmin)) * vm);
	server.print(sbuf);
	sprintf_P(sbuf, PSTR("<text x=\"%d\" y=\"%d\" font-size=\"12\" fontfam>%d</text>\n"), x - 25, yoff + (ytop - (i - vmin)) * vm , i);
	server.print(sbuf);
    }
}

void output_graph (WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
char sbuf[150];
int minutes, thisv, lastv;
int safei;
    scan_log();
    if(highest_temp < 120) { highest_temp = 130; }
    if(lowest_temp > 150) { lowest_temp = 70; }
    xleft = 130;
    hm = 1;
    xstep = 4;
    yoff = 50;
    vm = 10;
    gmargin = 2;
    // graphwidth = sizeof(run_log)/sizeof(LOGENTRY);
    xmax = xleft + xstep * graphwidth;
    ytop = highest_temp - lowest_temp;
    vsize = ytop * vm;	// size of the graph
    x = xleft;
    server.printP(htmlHead);
    server.printP(PSTR("<svg width=\"100%\" height=\"100%\" version=\"1.1\" xmlns=\"http://www.w3.org/2000/svg\">\n"));
    // ------------- start of graph ------------------
    sprintf_P(sbuf, PSTR("<text x=\"%d\" y=\"%d\" font-size=\"14\">%s</text>\n"), xleft / 2 + ((xmax - xleft) / 2), yoff - 30, "cheese temperature profile");
    server.print(sbuf);
    // print top border
    sprintf_P(sbuf, PSTR("<line x1=\"%d\" y1=\"%d\" x2=\"%d\" y2=\"%d\" style=\"stroke:rgb(30,30,30);stroke-width:3\"/>\n"), xleft - gmargin, yoff-gmargin, xmax + gmargin, yoff-gmargin);
    server.print(sbuf);
    // print bottom border
    sprintf_P(sbuf, PSTR("<line x1=\"%d\" y1=\"%d\" x2=\"%d\" y2=\"%d\" style=\"stroke:rgb(30,30,30);stroke-width:3\"/>\n"), xleft - gmargin, yoff + (ytop * vm) + gmargin, xmax + gmargin, yoff + (ytop * vm) + gmargin);
    server.print(sbuf);
    // print left border
    sprintf_P(sbuf, PSTR("<line x1=\"%d\" y1=\"%d\" x2=\"%d\" y2=\"%d\" style=\"stroke:rgb(30,30,30);stroke-width:3\"/>\n"), xleft - gmargin, yoff + 1 - gmargin, xleft - gmargin, yoff + (vm * ytop) + gmargin);
    server.print(sbuf);
    // print right border
    sprintf_P(sbuf, PSTR("<line x1=\"%d\" y1=\"%d\" x2=\"%d\" y2=\"%d\" style=\"stroke:rgb(30,30,30);stroke-width:3\"/>\n"), xmax + gmargin, yoff + 1 - gmargin, xmax + gmargin, yoff + (vm * ytop) + gmargin);
    server.print(sbuf);
    sprintf_P(sbuf, PSTR("<text x=\"%d\" y=\"%d\" font-size=\"12\" fontfam>%s</text>\n"), xleft - 50, yoff + (ytop /2) * vm , "&deg;F");
    server.print(sbuf);
    y_lines(server, yoff, xleft - gmargin, xmax + gmargin, vsize, lowest_temp, highest_temp);
    minutes = 1;
    safei = 1;
    for(uint8_t i = 1 ; i < config_copy.log_index; i++) {
#if MAXLOGENTRY < 256
	if(i >= MAXLOGENTRY) { i = 0; }
#endif
	if(i == config_copy.log_index) { break; }
	for(int graph = 0 ; graph < 3 ; graph++) {
	    switch (graph) {
		case 0 :
		    thisv = run_log[i].set_temp;
		    lastv = run_log[i-1].set_temp;
		    break;
		case 1 :
		    thisv = run_log[i].set_temp_i;
		    lastv = run_log[i-1].set_temp_i;
		    break;
		case 2 :
		    thisv = run_log[i].tempF;
		    lastv = run_log[i-1].tempF;
		    break;
	    }
	    if(thisv == 0 && lastv > 0) { thisv = lastv; }
	    if(lastv == 0) { lastv = thisv; }
	    sprintf_P(sbuf, PSTR("<line x1=\"%d\" y1=\"%d\" x2=\"%d\" y2=\"%d\" style=\"stroke:rgb(%s);stroke-width:2\"/>\n"), x, yoff + (ytop - (lastv - lowest_temp)) * vm, x + xstep, yoff + (ytop - (thisv - lowest_temp)) * vm, graph2color[graph]);
	    server.print(sbuf);
	}
	if((minutes % 10) == 0) {
	    sprintf_P(sbuf, PSTR("<text x=\"%d\" y=\"%d\" font-size=\"10\">%d</text>\n"), x, yoff + (ytop * vm) + 20, minutes);
	    server.print(sbuf);
	}
	x += xstep;
	minutes++;
	if(++safei >= MAXLOGENTRY) { break; }
    }
    sprintf_P(sbuf, PSTR("<text x=\"%d\" y=\"%d\" font-size=\"11\">%s</text>\n"), xleft, yoff + (ytop * vm) + 20, "minutes");
    server.print(sbuf);
    sprintf_P(sbuf, PSTR("<text x=\"%d\" y=\"%d\" fill=\"%s\" font-size=\"14\">%s</text>\n"), xleft + (10 * xstep), yoff + (ytop * vm) + 40, "red", "red = actual temp");
    server.print(sbuf);
    sprintf_P(sbuf, PSTR("<text x=\"%d\" y=\"%d\" fill=\"%s\" font-size=\"14\">%s</text>\n"), xleft + (60 * xstep), yoff + (ytop * vm) + 40, "blue", "blue = set point");
    server.print(sbuf);
    sprintf_P(sbuf, PSTR("<text x=\"%d\" y=\"%d\" fill=\"%s\" font-size=\"14\">%s</text>\n"), xleft + (110 * xstep), yoff + (ytop * vm) + 40, "green", "green = rate limited set point");
    server.print(sbuf);
    // ------------- end of graph ----------------------
    server.printP(PSTR("</svg>\n"));
    server.printP(root_link);
    server.printP(htmlTail);
}
